-- Functions used to create the front and up vectors for the controls   
-- Only the one used are exported
-- © Riccardo Di Meo 2015

local function vec_product(a, b) 
return {
   a[2] * b[3] - a[3] * b[2],
 -(a[1] * b[3] - a[3] * b[1]),
   a[1] * b[2] - a[2] * b[1]
}
end			

local function grad2rad(angle)
   return angle * math.pi / 180 
end

local function rad2grad(angle)
   return 180 * angle / math.pi 
end

local function length(v)
   return math.sqrt(v[1] * v[1] + v[2] * v[2] + v[3] * v[3])
end

local function normalize(v) 
   local l =  length(v)
   if l == 0 then
      return v
   end
   return {
      v[1] / l, 
      v[2] / l, 
      v[3] / l
   }
end

local function front_vector_from_angles(yaw, pitch) 
   local ryaw = grad2rad(yaw + 90)
   local rpitch = grad2rad(90 - pitch)
   return {
      math.cos(ryaw) * math.sin(rpitch),
                       math.cos(rpitch),
      (-math.sin(ryaw) * math.sin(rpitch))
   }
end

local function addv(a, b) 
   return {
      a[1] + b[1],
      a[2] + b[2],
      a[3] + b[3]
   }
end


local function length_3D(vec) 
   assert(#vec == 3)

   return math.sqrt(vec[1] * vec[1] + vec[2] * vec[2] + vec[3] * vec[3])
end

local function normalize_3D(vec)
   assert(#vec == 3)
   
   local l = length_3D(vec)
   
   if l > 0 then
      return {vec[1] / l, vec[2] / l, vec[3] / l}
   else
      return vec
   end
   
end

local function dot_3D(a, b)
   assert(#a == 3 and #b == 3)

   return a[1] * b[1] + a[2] * b[2] + a[3] * b[3]
end

local function mult_3D(a, b)
   assert(#a == 3 and #b == 3)

   local res = {
      a[2] * b[3] - a[3] * b[2], 
      a[3] * b[1] - a[1] * b[3], 
      a[1] * b[2] - a[2] * b[1]
   }

   assert(dot_3D(res, a) < 0.001)
   assert(dot_3D(res, b) < 0.001)

   return res
end

local function add_3D(a, b)
   assert(#a == 3 and #b == 3)
   return {
      a[1] + b[1],
      a[2] + b[2],
      a[3] + b[3]
   }
end

local function yaw_3D(vec, angle)
   local c = math.cos(angle)
   local s = math.sin(angle)
   return {
      c * vec[1] - s * vec[3],
      vec[2],
      s * vec[1] + c * vec[3]
   }
end

local function pitch_3D(vec, angle)
   local c = math.cos(angle)
   local s = math.sin(angle)
   return {
      vec[1],
      c * vec[2] - s * vec[3],
      s * vec[2] + c * vec[3]
   }
end

local function repr_3D(v)
   assert(#v == 3)
   
   return "(" .. v[1] .. ", " .. v[2] .. ", " .. v[3] .. ")"
end

local function smul_3D(a, x)
   assert(#a == 3)
   return {a[1] * x, a[2] * x, a[3] * x}
end

local function neg_3D(a)
   assert(#a == 3)
   return smul_3D(a, -1)
end

local function sub_3D(a, b)
   assert(#a == 3 and #b == 3)
   return add_3D(a, neg_3D(b))
end

local function copy_3D(a)
   assert(#a == 3)
   return {a[1], a[2], a[3]}
end


local function to_string_3D(a)
   return "|" .. tostring(a[1]) .. ", " .. tostring(a[2]) .. ", " ..  tostring(a[3]) .. "|"
end

local M = {}
M.vec_product = vec_product
M.grad2rad = grad2rad
M.rad2grad = rad2grad
M.front_vector_from_angles = front_vector_from_angles
M.add_3D = add_3D
M.sub_3D = sub_3D
M.copy_3D = copy_3D
M.smul_3D = smul_3D
M.neg_3D = neg_3D
M.repr_3D = repr_3D
M.length_3D = length_3D
M.normalize_3D = normalize_3D
M.mult_3D = mult_3D
M.dot_3D = dot_3D
M.pitch_3D = pitch_3D
M.yaw_3D = yaw_3D
M.to_string_3D = to_string_3D
return M
