-- Controls creation and configuration for the scene
-- (both plain and panoramic)
-- Created by Riccardo Di Meo
-- © Riccardo Di Meo 2015
local consts = require "scene/scene_constants"
local options = require "options"
-- TODO/FIXME: use the constants in this file!
local options_consts = require "scene/options_consts"
local scene_objects = require "scene/controls/scene_objects"
local scene_common = require "scene/scene_common"
local vr_config = require "scene/vr_config"
local controls_common = require "scene/controls/common"

-- Not dry with panoramic.lua
local set_all_controls_visibility = function(objects, shown)
    local all_controls = {
        consts.CONTROLS_PANEL, consts.CONTROLS_BUTTON,
        consts.INVERT_3D, consts.BACKWARD, consts.FORWARD,
        consts.VOLUME_UP, consts.VOLUME_DOWN,
        consts.LOCK_THEATER, consts.PAUSE, consts.ALIGN,
        consts.FAST_BACKWARD, consts.FAST_FORWARD,
        consts.MOVE_FORWARD, consts.MOVE_BACKWARD,
        consts.PREV_TRACK, consts.NEXT_TRACK,
        consts.MOVE_UP, consts.MOVE_DOWN, consts.ZOOM_IN, consts.ZOOM_OUT,
        consts.RESET_ZOOM, consts.SCREEN_TYPE, consts.SCREEN_RATIO,
        consts.STEREO_TYPE,
    }

    local dirty_check = false

    -- sloooowwww
    for _, name in ipairs(all_controls) do
        local index, obj = scene_common.get_scene_object_named(objects, name)
        -- hide everything, regardless!
        if index then
            obj.visible = false
            obj.dirty = true
            dirty_check = true
        end
    end
    -- and only if the controls are supposed to be shown, display the menù opener
    local index, dabutton = scene_common.get_scene_object_named(objects, consts.CONTROLS_BUTTON)
    if index then
        dabutton.dirty = true
        dirty_check = true
        dabutton.visible = shown and controls_common.get_controls_visibility()
    else
        error("No button???? C'mon!")
    end

    if dirty_check then controls_common.turn_dirty_check_on() end
end

-- Generic callbacks implementations
local generate_set_progress_visibility = function(objects)
    return function(visibility)
        local index, obj = scene_common.get_scene_object_named(objects, consts.PROGRESS)
        obj.visible = visibility
        obj.dirty = true
        controls_common.turn_dirty_check_on()
    end
end

local generate_started_fast_movement = function(objects)
   return function()
      controls_common.set_controls_visibility(false)
      set_all_controls_visibility(objects, false)
      local index, resume = scene_common.get_scene_object_named(objects, consts.RESUME)
      assert(resume ~= nil)
      resume.visible = true
      resume.dirty = true
      controls_common.turn_dirty_check_on()
    end
end

local generate_ended_fast_movement = function(objects)
    return function()
      controls_common.set_controls_visibility(true)
      set_all_controls_visibility(objects, true)   
      local index, resume = scene_common.get_scene_object_named(objects, consts.RESUME)
      assert(resume ~= nil)
      resume.visible = false
      resume.dirty = true
      controls_common.turn_dirty_check_on()
    end
end

local generate_alignment_countdown_event = function(objects, is_started)
   return function()
      controls_common.set_controls_visibility(not is_started)
      set_all_controls_visibility(objects, not is_started)
      local index, progress_bar = scene_common.get_scene_object_named(objects, consts.PROGRESS)
      progress_bar.visible = is_started
      progress_bar.dirty = true
      controls_common.turn_dirty_check_on()
    end
end

local generate_alignment_countdown_started = function(objects)
   return generate_alignment_countdown_event(objects, true)
end

local generate_alignment_countdown_ended = function(objects)
   return generate_alignment_countdown_event(objects, false)
end

-- scene callbacks not associated with a specific object: used to intercept events 
-- lua would not be privy about
local function generate_generic_callbacks(objects)
    local result = {}

    -- A keyboard action has been activated
    result.keyboard_action_executed = nil -- function(action_name)

    -- The user "clicked" on an inactive object
    result.trigger_on_empty_executed = nil -- function()

    -- a progress bar of sort should be displayed
    result.set_progress_visibility = generate_set_progress_visibility(objects)

    -- The user initiated a fast-forward/backward
    result.started_fast_movement = generate_started_fast_movement(objects)

    -- In a way or the other, a fast-forward/backward has been interrupted
    result.ended_fast_movement = generate_ended_fast_movement(objects)

    result.alignment_countdown_started = generate_alignment_countdown_started(objects)

    result.alignment_countdown_ended = generate_alignment_countdown_ended(objects)

    return result
end

local function set_scene_object_visibility(name, obj_list, visible)
   for i, obj in pairs(obj_list) do
      if obj.name == name then
         obj.visible = visible
         obj.dirty = true
      end
   end                
end


local function set_pointer_visibility(obj_list, visible)
   set_scene_object_visibility(consts.POINTER, obj_list, visible)
end

local function set_seek_bar_visibility(obj_list, visible)
   set_scene_object_visibility(consts.SEEK_BAR, obj_list, visible)
end

local function generate_screen_focus_handler(obj_list, is_start)
   local auto_hide = vr_config.get(options_consts.SEEK_BAR_AUTO_HIDE)
   local pointer_always_on = vr_config.get(options_consts.POINTER_ALWAYS_ON)
   return {
      post_native = function(prev_object)
         if not pointer_always_on then
            set_pointer_visibility(obj_list, not is_start)
         end
         if auto_hide then 
            set_seek_bar_visibility(obj_list, not is_start)
         end
         controls_common.turn_dirty_check_on()
      end
   }
end

local function generate_screen_start_focus(obj_list)
   return generate_screen_focus_handler(obj_list, true)
end

local function generate_screen_end_focus(obj_list)
   return generate_screen_focus_handler(obj_list, false)
end

return {
   create_theater_controls = scene_objects.create_theater_controls,
   create_panoramic_controls = scene_objects.create_panoramic_controls,
   generate_generic_callbacks = generate_generic_callbacks,
   generate_screen_start_focus = generate_screen_start_focus,
   generate_screen_end_focus = generate_screen_end_focus,
   create_seek_bar = scene_objects.create_seek_bar,
   create_seek_bar_cubemap = scene_objects.create_seek_bar_cubemap,
}
