local actions_consts = require "scene/actions_consts"
local consts = require "scene/scene_constants"
local controls_common = require "scene/controls/common"
local controls_sets = require "scene/controls/sets"
local options_consts = require "scene/options_consts"
local quat = require "scene/quaternion"
local scene_objects_creation = require "scene/controls/scene_objects_creation"

local function create_theater_controls(camera_obj, ui, buttons_distance,
                                       pitch, yaw, callbacks)
   
   local movement_function = callbacks.movement
   local save_config_function = callbacks.config_save
   
   -- all the components, including reference scene objects
   local scene_objects = scene_objects_creation.new_objects_table(
      camera_obj.position, ui.distance,
      pitch, yaw)

   scene_objects.panel.action = {
      end_focus = scene_objects:handle_end_focus()
   }

   scene_objects.button.action = {
      start_focus = {
         pre_native = function(prev_object)
            scene_objects:turn_controls_on()
         end
      },
   }

   scene_objects.switch_button.action = {
      start_focus = {
         pre_native = function()
            scene_objects:change_controls_set()
         end
      },
      end_focus = scene_objects:handle_end_focus()
   }

   local function wrap_native_action(native_action, repeatable)
      local select_action = {
         native = native_action,
         post_native = function(prev_object)
            if not repeatable then
               scene_objects:turn_controls_off()
            end
         end
      }

      -- Create a suitable action from a native action
      return {
         end_focus = scene_objects:handle_end_focus(),
         trigger = select_action,
         vselected = select_action,
      }
   end

   scene_objects:create_theater_buttons(buttons_distance, ui.button_side)

   -- 1th panel
   scene_objects.all_buttons.pause.action = wrap_native_action(actions_consts.ACTION_PAUSE, false)
   scene_objects.all_buttons.forward.action = wrap_native_action(actions_consts.ACTION_FORWARD, true)
   scene_objects.all_buttons.align.action = wrap_native_action(actions_consts.ACTION_ALIGN_COUNTER, false)
   scene_objects.all_buttons.fast_forward.action = wrap_native_action(actions_consts.ACTION_START_FAST_FORWARD, false)
   scene_objects.all_buttons.next_track.action = wrap_native_action(actions_consts.ACTION_NEXT_TRACK, false)
   scene_objects.all_buttons.prev_track.action = wrap_native_action(actions_consts.ACTION_PREV_TRACK, false)
   scene_objects.all_buttons.fast_backward.action = wrap_native_action(actions_consts.ACTION_START_FAST_BACKWARD, false)
   scene_objects.all_buttons.volume_up.action = wrap_native_action(actions_consts.ACTION_VOLUME_UP, true)
   scene_objects.all_buttons.volume_down.action = wrap_native_action(actions_consts.ACTION_VOLUME_DOWN, true)
   scene_objects.all_buttons.backward.action = wrap_native_action(actions_consts.ACTION_BACKWARD, true)  

   -- 2th panel
   scene_objects.all_buttons.lock_theater.action = wrap_native_action(actions_consts.ACTION_LOCK_VIEW, false)    
   scene_objects.all_buttons.invert_3d.action = wrap_native_action(actions_consts.ACTION_INVERT_3D, false)    
   scene_objects.all_buttons.stereo_type.action = wrap_native_action(actions_consts.ACTION_NEXT_STEREO_TYPE, false)    
   scene_objects.all_buttons.screen_type.action = wrap_native_action(actions_consts.ACTION_CHANGE_SCREEN_TYPE, false)    
   scene_objects.all_buttons.screen_ratio.action = wrap_native_action(actions_consts.ACTION_NEXT_SCREEN_RATIO, false)    

   -- used to move the theater (or, better, the camera) in non panoramic views
   local displacement_action = function(displacement)
      local hef = scene_objects:handle_end_focus()
      local old_pre_native = hef.pre_native

      hef.pre_native = function(next_object)
         old_pre_native(next_object)
         if save_config_function then save_config_function() end
      end

      return {
         end_focus = hef,
         in_focus = {
            pre_native = function()
               if movement_function(displacement) then
                  scene_objects.controls_group.position = scene_objects_creation.controls_group_position(
                     camera_obj.position, ui.distance,
                     pitch, yaw)
                  
                  scene_objects.controls_group.dirty = true

                  controls_common.turn_dirty_check_on()
               end
            end
         }
      }
   end

   scene_objects.all_buttons.move_forward.action = displacement_action({ 0.0, 0.0, 0.025 })
   scene_objects.all_buttons.move_backward.action = displacement_action({ 0.0, 0.0, -0.025 })
   scene_objects.all_buttons.move_up.action = displacement_action({ 0.0, -0.025, 0.0 })        
   scene_objects.all_buttons.move_down.action = displacement_action({ 0.0, 0.025, 0.0 })        

   scene_objects:turn_controls_off()

   return scene_objects:as_list()
end

local function create_panoramic_controls(camera_obj, ui, buttons_distance,
                                         pitch, yaw, callbacks)
   
   local zoom_function = callbacks.zooming
   local save_config_function = callbacks.config_save
   
   -- all the components, including reference scene objects
   local scene_objects = scene_objects_creation.new_objects_table(
      camera_obj.position, ui.distance,
      pitch, yaw)

   scene_objects.panel.action = {
      end_focus = scene_objects:handle_end_focus()
   }

   scene_objects.button.action = {
      start_focus = {
         pre_native = function(prev_object)
            scene_objects:turn_controls_on()
         end
      },
   }

   scene_objects.switch_button.action = {
      start_focus = {
         pre_native = function()
            scene_objects:change_controls_set()
         end
      },
      end_focus = scene_objects:handle_end_focus()
   }

   local function wrap_native_action(native_action, repeatable)
      local select_action = {
         native = native_action,
         post_native = function(prev_object)
            if not repeatable then
               scene_objects:turn_controls_off()
            end
         end
      }

      -- Create a suitable action from a native action
      return {
         end_focus = scene_objects:handle_end_focus(),
         trigger = select_action,
         vselected = select_action,
      }
   end

   scene_objects:create_panoramic_buttons(buttons_distance, ui.button_side)

   -- 1th panel
   scene_objects.all_buttons.pause.action = wrap_native_action(actions_consts.ACTION_PAUSE, false)
   scene_objects.all_buttons.forward.action = wrap_native_action(actions_consts.ACTION_FORWARD, true)
   scene_objects.all_buttons.align.action = wrap_native_action(actions_consts.ACTION_ALIGN_COUNTER, false)
   scene_objects.all_buttons.fast_forward.action = wrap_native_action(actions_consts.ACTION_START_FAST_FORWARD, false)
   scene_objects.all_buttons.next_track.action = wrap_native_action(actions_consts.ACTION_NEXT_TRACK, false)
   scene_objects.all_buttons.prev_track.action = wrap_native_action(actions_consts.ACTION_PREV_TRACK, false)
   scene_objects.all_buttons.fast_backward.action = wrap_native_action(actions_consts.ACTION_START_FAST_BACKWARD, false)
   scene_objects.all_buttons.volume_up.action = wrap_native_action(actions_consts.ACTION_VOLUME_UP, true)
   scene_objects.all_buttons.volume_down.action = wrap_native_action(actions_consts.ACTION_VOLUME_DOWN, true)
   scene_objects.all_buttons.backward.action = wrap_native_action(actions_consts.ACTION_BACKWARD, true)  

   -- 2th panel
   scene_objects.all_buttons.lock_theater.action = wrap_native_action(actions_consts.ACTION_LOCK_VIEW, false)    
   scene_objects.all_buttons.invert_3d.action = wrap_native_action(actions_consts.ACTION_INVERT_3D, false)    
   scene_objects.all_buttons.stereo_type.action = wrap_native_action(actions_consts.ACTION_NEXT_STEREO_TYPE, false)    
   scene_objects.all_buttons.screen_type.action = wrap_native_action(actions_consts.ACTION_CHANGE_SCREEN_TYPE, false)    
   scene_objects.all_buttons.screen_ratio.action = wrap_native_action(actions_consts.ACTION_NEXT_SCREEN_RATIO, false)    

   -- used to move the theater (or, better, the camera) in non panoramic views
   local zoom_action = function(zoom_factor)
      local hef = scene_objects:handle_end_focus()
      local old_pre_native = hef.pre_native
      
      hef.pre_native = function(next_object)
         old_pre_native(next_object)
         if save_config_function then save_config_function() end
      end
      
      return {
         end_focus = hef,
         in_focus = {
            pre_native = function()
               if zoom_function(zoom_factor) then
                  controls_common.turn_dirty_check_on()
               end
            end
         }
      }
   end
   
   local function reset_zoom()
      local select_action = {
         native = actions_consts.ACTION_VOID,
         post_native = function(prev_object)
            if zoom_function() then
               controls_common.turn_dirty_check_on()
               save_config_function()
            end
         end
      }
      
      -- Create a suitable action from a native action
      return {
         end_focus = scene_objects:handle_end_focus(),
         trigger = select_action,
         vselected = select_action,
      }
   end
   
   scene_objects.all_buttons.zoom_in.action = zoom_action(0.5)
   scene_objects.all_buttons.zoom_out.action = zoom_action(-0.5)
   scene_objects.all_buttons.reset_zoom.action = reset_zoom()

   scene_objects:turn_controls_off()

   return scene_objects:as_list()
end

function create_seek_bar(screen_params, shown, seek_bar_position)
   local position
   local scale
   
   if seek_bar_position == options_consts.SEEK_BAR_POSITION_TOP then
      position = {0, screen_params.height * 0.5 + 0.6, 0}
      scale = {screen_params.width, 0.8, 1}
   elseif seek_bar_position == options_consts.SEEK_BAR_POSITION_BOTTOM then
      position = {0, -screen_params.height * 0.5 - 0.6, 0}
      scale = {screen_params.width, 0.8, 1 }
   elseif seek_bar_position == options_consts.SEEK_BAR_POSITION_LEFT then
      position = {-screen_params.width * 0.5 - 0.6, 0, 0}
      scale = {0.8, screen_params.height, 1}
   elseif seek_bar_position == options_consts.SEEK_BAR_POSITION_RIGHT then
      position = {screen_params.width * 0.5 + 0.6, 0, 0}
      scale = {0.8, screen_params.height, 1}
   end

   local scene_object = {
      name = consts.SEEK_BAR,
      position = position,
      scale = scale,
      orientation = quat.front_up_to_quaternion(
         { 0, 0, 1 },
         { 0, 1, 0 }),
      relativeto = consts.SCREEN_PLACEMENT,
      action = {
         trigger = {
            native = actions_consts.ACTION_SEEK,
         },
         vselected = {
            native = actions_consts.ACTION_SEEK,
         }
      }
   }

   if shown then
      scene_object.openglmodel = "current.opengl.seekbar"
   end

   return scene_object
end

function create_seek_bar_cubemap(screen_params, shown, seek_bar_position)
   local position
   local scale
   
   if seek_bar_position == options_consts.SEEK_BAR_POSITION_TOP then
      position = {0, screen_params.height * 0.5 + 3, 0}
      scale = {screen_params.width, 4, 1}
   elseif seek_bar_position == options_consts.SEEK_BAR_POSITION_BOTTOM then
      position = {0, -screen_params.height * 0.5 - 3, 0}
      scale = {screen_params.width, 4, 1 }      
   elseif seek_bar_position == options_consts.SEEK_BAR_POSITION_LEFT then
      position = {-screen_params.width * 0.5 - 3, 0, 0}
      scale = {4, screen_params.height, 1}      
   elseif seek_bar_position == options_consts.SEEK_BAR_POSITION_RIGHT then
      position = {screen_params.width * 0.5 + 3, 0, 0}
      scale = {4, screen_params.height, 1}      
   end

   local scene_object = {
      name = consts.SEEK_BAR,
      position = position,
      scale = scale,
      orientation = quat.front_up_to_quaternion(
         { 0, 0, -1 },
         { 0, 1, 0 }),
      relativeto = consts.SCREENGROUP,
      action = {
         trigger = {
            native = actions_consts.ACTION_SEEK,
         },
         vselected = {
            native = actions_consts.ACTION_SEEK,
         }
      }
   }

   if shown then
      scene_object.openglmodel = "current.opengl.seekbar"
   end

   return scene_object
end


return {
   create_theater_controls = create_theater_controls,
   create_panoramic_controls = create_panoramic_controls,
   create_seek_bar = create_seek_bar,
   create_seek_bar_cubemap = create_seek_bar_cubemap,
}
