local consts = require "scene/scene_constants"
local controls_common = require "scene/controls/common"
local controls_sets = require "scene/controls/sets"
local mymath = require "mymath"
local quat = require "scene/quaternion"
local options = require "options"

local function create_panel()
   local panel = {}
   panel.name = consts.CONTROLS_PANEL
   panel.relativeto = consts.CONTROLS_BASE
   panel.openglmodel = "current.opengl.controls_panel"
   panel.position = { 0, 0, 0 }
   panel.orientation = quat.front_up_to_quaternion(
      { 0, 0, 1 }, { 0, 1, 0 })
   return panel
end

local function viewer_pos_world_coordinates(camera_pos) 
   return mymath.neg_3D(camera_pos)
end

local controls_group_position = function(camera_pos, distance, pitch, yaw)
   local me = viewer_pos_world_coordinates(camera_pos)
   
   local front_of_viewer = {0, 0, (-distance)}
   local right_elevation = mymath.pitch_3D(front_of_viewer, pitch)
   -- relative end position for the controls:
   local end_controls_position = mymath.yaw_3D(right_elevation, yaw)

   return mymath.add_3D(me, end_controls_position)
end

local function create_controls_group(camera_pos, controls_distance, controls_pitch, controls_yaw)
   local controls_group = {}
   controls_group.name = consts.CONTROLS_GROUP
   controls_group.position = controls_group_position(
      camera_pos, controls_distance, controls_pitch, controls_yaw)   
   controls_group.orientation = quat.front_up_to_quaternion(
      { 0, 0, -1 }, { 0, 1, 0 })
   return controls_group
end

local function create_red_button()
   local button = {}
   button.name = consts.CONTROLS_BUTTON
   button.relativeto = consts.CONTROLS_BASE
   button.openglmodel = "current.opengl.controls_button"
   button.position = { 0, 0, 0.1 }
   button.orientation = quat.front_up_to_quaternion({ 0, 0, 1 }, { 0, 1, 0 })
   return button
end

local function create_switch_button()
   local switch_button = {}
   switch_button.name = consts.CONTROLS_SWITCH
   switch_button.relativeto = consts.CONTROLS_BASE
   switch_button.openglmodel = "current.opengl.switch_button"
   switch_button.position = { -0.75, 0.75, 0 }
   switch_button.scale = { 0.66, 0.66, 1.0 }
   switch_button.orientation = quat.front_up_to_quaternion(
      { 0, 0, 1 }, { 0, 1, 0 })
   return switch_button
end

local function create_pointer() 
   local pointer = {}
   pointer.name = consts.POINTER
   pointer.openglmodel = "current.opengl.pointer"
   pointer.position = { 0, 0, -6 }
   pointer.orientation = quat.front_up_to_quaternion(
      { 0, 0, -1 }, { 0, 1, 0 })
    return pointer
end

local controls_group_orientation = function(pitch, yaw)
   -- In case it wasn't clear, I don't remember how I did this anymore…
   local up_vector = mymath.pitch_3D({0, 1, 0}, (-pitch))
   local mistery = mymath.yaw_3D({0, 0, 1}, yaw)
   local front_vector = mymath.pitch_3D(mistery, (-pitch))
   return quat.front_up_to_quaternion(front_vector, up_vector)
end

local function create_controls_reference(pitch, yaw)
    local controls_reference = {}
    controls_reference.name = consts.CONTROLS_BASE
    controls_reference.relativeto = consts.CONTROLS_GROUP
    controls_reference.position = { 0, 0, 0 }
    controls_reference.orientation = controls_group_orientation(pitch, yaw)
    return controls_reference
end

local function set_controls_visibility(scene_objects, controls_visible)
   scene_objects.button.visible = not controls_visible
   scene_objects.button.dirty = true

   scene_objects.panel.visible = controls_visible
   scene_objects.panel.dirty = true

   scene_objects.switch_button.visible = controls_visible
   scene_objects.switch_button.dirty = true

   for i, j in pairs(scene_objects.all_buttons) do
      j.visible = controls_visible and (controls_sets.get(j.name) ~= nil)
      j.dirty = true
   end

   if not controls_common.get_controls_visibility() then
      scene_objects.button.visible = false
   end

   controls_common.turn_dirty_check_on()
end

-- TODO: lies about its name?
local function turn_controls_on(self)
   if controls_common.get_controls_visibility() then
      -- Reset the controls that are to be shown
      controls_sets.set_first_panel()
      self:set_controls_visibility(true)
   end
end

local function turn_controls_off(self)
   self:set_controls_visibility(false)
end

local function handle_end_focus(self)
   return {
      pre_native = function(next_object)
         if next_object == nil or not controls_sets.is_a_control(next_object) then
            self:turn_controls_off()
         end
      end
   }
end

local change_controls_set = function(self)
   controls_sets.switch_controls_panel()
   self:set_controls_visibility(true)
end

local function get_button_coordinates(index, slots, distance, yshift)
   local angle = math.pi * (0.5 - index * 2.0 / slots)
   return {
      math.cos(angle) * distance,
      math.sin(angle) * distance + (yshift or 0),
      0
   }
end

local function generate_button(name, model, index, slots, distance, yshift)
   local button = {}
   button.name = name
   button.relativeto = consts.CONTROLS_BUTTON
   button.openglmodel = model
   button.position = get_button_coordinates(index, slots, distance, yshift)
   button.orientation = quat.front_up_to_quaternion({ 0, 0, -1 }, { 0, 1, 0 })
   return button
end

local function curried_generate_button(slots, distance)
   return function(name, model, index, yshift)
      return generate_button(name, model, index, slots, distance, yshift)
   end
end

local function create_common_buttons(self, create_button, button_side)
    -- 1th panel
    self.all_buttons.pause = create_button(consts.PAUSE, "current.opengl.pause", 0)
    self.all_buttons.forward = create_button(consts.FORWARD, "current.opengl.forward", 1)
    self.all_buttons.align = create_button(consts.ALIGN, "current.opengl.align", 2)
    self.all_buttons.fast_forward = create_button(consts.FAST_FORWARD, "current.opengl.fast_forward", 3)
    self.all_buttons.next_track = create_button(consts.NEXT_TRACK, "current.opengl.next_track", 4)
    self.all_buttons.prev_track = create_button(consts.PREV_TRACK, "current.opengl.prev_track", 5)
    self.all_buttons.fast_backward = create_button(consts.FAST_BACKWARD, "current.opengl.fast_backward", 6)
    self.all_buttons.volume_up = create_button(consts.VOLUME_UP, "current.opengl.volumeup", 7, button_side / 2)
    self.all_buttons.volume_down = create_button(consts.VOLUME_DOWN, "current.opengl.volumedown", 7, -button_side / 2)
    self.all_buttons.backward = create_button(consts.BACKWARD, "current.opengl.backward", 8)
    -- 2th panel
    self.all_buttons.lock_theater = create_button(consts.LOCK_THEATER, "current.opengl.lock", 9)
    self.all_buttons.invert_3d = create_button(consts.INVERT_3D, "current.opengl.invert", 8)
    self.all_buttons.stereo_type = create_button(consts.STEREO_TYPE, "current.opengl.stereo_type", 7)
    self.all_buttons.screen_type = create_button(consts.SCREEN_TYPE, "current.opengl.screen_type", 6)
    self.all_buttons.screen_ratio = create_button(consts.SCREEN_RATIO, "current.opengl.screen_ratio", 5)
end

local function create_theater_buttons(self, buttons_distance, button_side)
   local slots = 9
   local create_button = curried_generate_button(slots, buttons_distance)

   create_common_buttons(self, create_button, button_side)
   self.all_buttons.move_forward = create_button(consts.MOVE_FORWARD, "current.opengl.move_forward", 2, button_side / 2)
   self.all_buttons.move_backward = create_button(consts.MOVE_BACKWARD, "current.opengl.move_backward", 2, -button_side / 2)
   self.all_buttons.move_up = create_button(consts.MOVE_UP, "current.opengl.move_up", 1, button_side / 2)
   self.all_buttons.move_down = create_button(consts.MOVE_DOWN, "current.opengl.move_down", 1, -button_side / 2)
end

local function create_panoramic_buttons(self, buttons_distance, button_side)
   local slots = 9
   local create_button = curried_generate_button(slots, buttons_distance)

    create_common_buttons(self, create_button, button_side)
    self.all_buttons.zoom_in = create_button(consts.ZOOM_IN, "current.opengl.zoom_in", 1, button_side / 2)
    self.all_buttons.zoom_out = create_button(consts.ZOOM_OUT, "current.opengl.zoom_out", 1, -button_side / 2)
    self.all_buttons.reset_zoom = create_button(consts.RESET_ZOOM, "current.opengl.reset_zoom", 2)
end

local function new_objects_table(camera_pos, controls_distance, controls_pitch, controls_yaw)
   local objects = {
      -- the point in space the panel will be relative to
      controls_reference = create_controls_reference(controls_pitch, controls_yaw),
      -- the panel containing all controls
      panel = create_panel(),
      -- the point in space the controls will be relative to
      controls_group = create_controls_group(camera_pos, controls_distance, controls_pitch, controls_yaw),      
      -- the controls on/off button
      button = create_red_button(),
      -- the "more controls" button
      switch_button = create_switch_button(),
      -- the pointer
      pointer = create_pointer(),
      -- all surrounding buttons
      all_buttons = {
         lock = {},
      }
   }

   objects.set_controls_visibility = set_controls_visibility
   objects.turn_controls_off = turn_controls_off
   objects.turn_controls_on = turn_controls_on
   objects.handle_end_focus = handle_end_focus
   objects.change_controls_set = change_controls_set
   objects.create_theater_buttons = create_theater_buttons
   objects.create_panoramic_buttons = create_panoramic_buttons

   function objects:as_list()
      local result = {}
      result[#result + 1] = self.controls_group
      result[#result + 1] = self.controls_reference
      result[#result + 1] = self.panel
      result[#result + 1] = self.button
      result[#result + 1] = self.switch_button

      for _, j in pairs(self.all_buttons) do
         result[#result + 1] = j
      end

      result[#result + 1] = self.pointer

      return result
   end
   
   return objects
end

return {
   controls_group_position = controls_group_position,
   new_objects_table = new_objects_table
}
