local vr_config = require "scene/vr_config"
local options_consts = require "scene/options_consts"
local common_uniforms = require "scene/common_uniforms"
local consts = require "scene/scene_constants"
local opengl_common = require "scene/opengl_common"
local functional = require "scene/functional"
local subtitles = require "scene/subtitles"
local scene_controls_opengl = require "scene/scene_controls_opengl"

local cubemap_common = require "scene/cubemap/common"
local opengl = require "opengl"

local function set_opengl_near_far_parameters()
        opengl.znear = 0.1
        opengl.zfar = 500.0
end

--- Not DRY with user_options_template.lua
local function cubemap_split_selected(current_value)
    return string.match(current_value, "([^:]*):(.*)$")
end

local function compute_skybox_distance()
        return cubemap_common.SMALL_CUBEMAP_SIZE
end

local function create_skybox_globject()
    local size = 2 * compute_skybox_distance()

    local marshalled_skybox = vr_config.get(options_consts.SELECTED_CUBEMAP)

    local _, skybox_data = cubemap_split_selected(marshalled_skybox)

    local side = size * 0.5

    return {
        {
            type = "bind program",
            shader = {
                vert = "shaders/cubemap/cubemap.vert",
                frag = "shaders/cubemap/cubemap.frag",
            }
        },
        {
            -- uses the current shader
            type = "bind uniforms",
            uniforms = {
                projection = common_uniforms.projection,
                modelview = common_uniforms.modelview,
                cubemap = { content = "CUBETEX:1:" .. skybox_data, vtype = "int" },
                yaw = {
                    content = vr_config.get(options_consts.CUBEMAP_YAW_CORRECTION) * (math.pi / 180.0),
                    vtype = "float"
                },
            }
        },
        {
            -- TODO/FIXME: convert this horror with the scale parameter!
            type = "bind attribute",
            name = "pos",
            values = {
                (-side), (-side), (-side),
                side, (-side), (-side),
                side, (-side), side,
                (-side), (-side), side,
                (-side), side, (-side),
                side, side, (-side),
                side, side, side,
                (-side), side, side,
            },
            dim = 3
        },
        {
            type = "draw",
            primitive = "GL_TRIANGLES",
            elements = {
                0, 1, 3, 3, 1, 2, -- floor
                4, 7, 5, 7, 6, 5, -- ceiling
                3, 2, 7, 7, 2, 6, -- back
                0, 4, 1, 4, 5, 1, -- front
                0, 3, 4, 4, 3, 7, -- left
                5, 2, 1, 5, 6, 2, -- right
            },
        },
        {
            type = "bind attribute",
            name = "pos"
        }
    }
end

-- TODO/FIXME: not dry with theater_like_scene
local create_screen = function(parameters, screen_params, direct_rendering)

    set_opengl_near_far_parameters()

    local uniforms = {
        projection = common_uniforms.projection,
        modelview = common_uniforms.modelview,
        eye = common_uniforms.eye,
        id = common_uniforms.id,
        brightness = common_uniforms.brightness,
        contrast = common_uniforms.contrast,
        feedback_rendering = common_uniforms.feedback_rendering,
    }
    -- different shaders for different situations
    if direct_rendering then
        uniforms.video = common_uniforms.video
        uniforms.video_transform = common_uniforms.video_transform
    else
        uniforms.prevideo = common_uniforms.video
    end

    -- those two are used only when "anaglyph red/cyan" is selected
    if parameters.stereo_type == consts.ARG_STEREO_TYPE_ANAGLYPH_RED_CYAN then
        uniforms.swizzle_right = { content = opengl_common.anaglyph_conversion(consts.COLOR_RED), vtype = "mat3" }
        uniforms.swizzle_left = { content = opengl_common.anaglyph_conversion(consts.COLOR_CYAN), vtype = "mat3" }
    end

    return {
        {
            type = "bind program",
            shader = {
                vert = opengl_common.get_screen_vertex_shader(parameters.stereo_type, parameters.ratio, direct_rendering),
                frag = opengl_common.get_screen_fragment_shader(parameters.stereo_type, parameters.ratio, direct_rendering)
            }
        },
        {
            -- uses the current shader
            type = "bind uniforms",
            uniforms = uniforms
        },
        {
            type = "bind attribute",
            name = "pos",
            values = {
                (-screen_params.width) * 0.5, screen_params.height * 0.5, 0,
                (-screen_params.width) * 0.5, -screen_params.height * 0.5, 0,
                screen_params.width * 0.5, -screen_params.height * 0.5, 0,
                screen_params.width * 0.5, screen_params.height * 0.5, 0
            },
            dim = 3
        },
        {
            type = "bind attribute",
            name = "uv",
            values = {
                0, 1,
                0, 0,
                1, 0,
                1, 1
            },
            dim = 2
        },
        {
            type = "draw",
            primitive = "GL_TRIANGLES",
            elements = { 0, 1, 2, 0, 2, 3 },
        },
        {
            type = "bind attribute",
            name = "uv"
        },
        {
            type = "bind attribute",
            name = "pos"
        }
    }
end

local function create_seek_bar()
   -- duplicated code with theater/renderer.lua
   function seek_bar_fragment_shader(seek_bar_position)
      if seek_bar_position == options_consts.SEEK_BAR_POSITION_TOP or seek_bar_position == options_consts.SEEK_BAR_POSITION_BOTTOM then
         return "shaders/seekbar/seekbar_horiz.frag"
      else
         return "shaders/seekbar/seekbar_vert.frag"
      end
   end

   return {
      {
         type = "bind program",
         shader = {
            vert = "shaders/seekbar/seekbar.vert",
            frag = seek_bar_fragment_shader(vr_config.get(options_consts.SEEK_BAR_POSITION))
         }
      },
      {
         -- uses the current shader
         type = "bind uniforms",
         uniforms = {
            id = common_uniforms.id,
            movietime = common_uniforms.movietime,
            modelview = {
               content = "VAR:CONTROLS_MODELVIEW",
               vtype = "mat4",
            },
            opacity = {
               content = (vr_config.get(options_consts.SEEK_BAR_OPACITY) / 100),
               vtype = "float"
            },
            projection = common_uniforms.projection,
            feedback_rendering = common_uniforms.feedback_rendering,
         }
      },
      {
         type = "bind attribute",
         name = "pos",
         values = {
            (-0.5),  0.5, 0,
            (-0.5),  (-0.5), 0,
            0.5,     (-0.5), 0,
            0.5,     0.5, 0
         },
         dim = 3
      },
      {
         type = "bind attribute",
         name = "uv",
         values = {
            0, 1,
            0, 0,
            1, 0,
            1, 1
         },
         dim = 2
      },
      {
         type = "draw",
         primitive = "GL_TRIANGLES",
         elements = { 0, 1, 2, 0, 2, 3 },
      },
      {
         type = "bind attribute",
         name = "uv"
      },
      {
         type = "bind attribute",
         name = "pos"
      }
   }
end

local function create_models(parameters, screen_params, screen_points)

    local direct_rendering = parameters.content_type ~= "image"

    -- OpenGL models
    local opengl_models = {}
    -- seek bar
    opengl_models.seekbar = create_seek_bar()
    -- screen
    opengl_models.screen = create_screen(parameters, screen_params, direct_rendering)
    -- controls
    opengl_models = functional.merge_maps(opengl_models, scene_controls_opengl.generate_control_objects(screen_params.height / 4, true, false, parameters.ratio, parameters.stereo_type))

    -- disable the depth test on the resume button
    opengl_models.resume.options = { ignore_depth = true }
    opengl_models.pointer = scene_controls_opengl.return_pointer(0.1,
        vr_config.get(options_consts.POINTER_OPACITY) / 100.0)

    local function get_subtitles_quad(screen_points)
        local sub_width, sub_height = subtitles.compute_size(vr_config.get(options_consts.SUBTITLES_QUAD_DISTANCE),
            vr_config.get(options_consts.SUBTITLES_QUAD_SIZE) / 100.0,
            screen_points)

        return {
            {
                type = "bind program",
                shader = {
                    vert = "shaders/subtitles/subtitles.vert",
                    frag = "shaders/subtitles/subtitles.frag"
                }
            },
            {
                -- uses the current shader
                type = "bind uniforms",
                uniforms = {
                    projection = common_uniforms.projection,
                    modelview = common_uniforms.modelview,
                    subtitle = common_uniforms.subtitle,
                }
            },
            {
                -- TODO/FIXME: convert this to the scale parameter
                type = "bind attribute",
                name = "pos",
                values = {
                    (-sub_width * 0.5), sub_height, 0,
                    (-sub_width * 0.5), 0, 0,
                    sub_width * 0.5, 0, 0,
                    sub_width * 0.5, sub_height, 0,
                },
                dim = 3
            },
            {
                type = "bind attribute",
                name = "uv",
                values = {
                    0, 0,
                    0, 1,
                    1, 1,
                    1, 0
                },
                dim = 2
            },
            {
                type = "disable",
                name = "GL_DEPTH_TEST"
            },
            {
                type = "draw",
                primitive = "GL_TRIANGLES",
                elements = { 0, 1, 2, 0, 2, 3 },
            },
            {
                type = "enable",
                name = "GL_DEPTH_TEST"
            },
            {
                type = "bind attribute",
                name = "uv"
            },
            {
                type = "bind attribute",
                name = "pos"
            }
        }
    end

    opengl_models.subtitles_quad = get_subtitles_quad(screen_points)

    opengl_models.progress = opengl_common.progress_bar_model(common_uniforms)

    local skybox_models = create_skybox_globject()

    -- cubemap walls
    return functional.merge_maps(opengl_models, {skybox = skybox_models})
end

return {
    create_models = create_models
}
