-- 3D scene details for cubemaps
-- Created by Riccardo Di Meo
-- © Riccardo Di Meo 2015
local actions_consts = require "scene/actions_consts"
local cardboardtheater = require "it.couchgames.lua.cardboardtheater"
local consts = require "scene/scene_constants"
local cubemap_common = require "scene/cubemap/common"
local functional = require "scene/functional"
local mymath = require "mymath"
local options = require "options"
local options_consts = require "scene/options_consts"
local quat = require "scene/quaternion"
local rendering = require "scene/cubemap/rendering"
local scene_controls = require "scene/controls/controls"
local subtitles = require "scene/subtitles"
local vr_config = require "scene/vr_config"

local get_screen_dimensions = function(diagonal, aspect_ratio)
    local height = diagonal / math.sqrt(1 + aspect_ratio * aspect_ratio)
    local width = height * aspect_ratio
    return { width, height }
end

local get_screen_parameters = function(aspect_ratio)
    local distance = vr_config.get(options_consts.CUBEMAP_SCREEN_DISTANCE)
    local lift = vr_config.get(options_consts.CUBEMAP_SCREEN_LIFT)
    local diagonal = vr_config.get(options_consts.CUBEMAP_SCREEN_DIAGONAL)

    local screen_dimensions = get_screen_dimensions(diagonal, aspect_ratio)

    return {
        distance = distance,
        lift = lift,
        width = screen_dimensions[1],
        height = screen_dimensions[2]
    }
end

local create_skybox = function(camera_obj)
    local scene_obj = {
        name = consts.SKYBOX,
        position = mymath.neg_3D(camera_obj.position),
        orientation = quat.front_up_to_quaternion({ 0, 0, -1 }, { 0, 1, 0 }),
        openglmodel = "current.opengl.skybox",
    }
    return scene_obj
end

local function create(parameters)
    local result = {}

    -- objects in the scene (from a logic/position point of view)
    result.scene = {}
    result.scene.player = consts.CAMERA
    result.scene.locked = consts.SCREEN

    result.scene.objects = {}

    result.scene.objects[#result.scene.objects + 1] = {
        name = consts.ROOM,
        position = { 0, 0, 0 },
        orientation = quat.front_up_to_quaternion({ 0, 0, -1 }, { 0, 1, 0 }),
    }

    -- Screen group
    local screen_params = get_screen_parameters(parameters.ratio)
    local screen_group_obj = {
        name = consts.SCREENGROUP,
        position = { 0, 0, 0 },
        orientation = quat.front_up_to_quaternion({ 0, 0, -1 }, { 0, 1, 0 }),
        relativeto = consts.ROOM,
    }
    result.scene.objects[#result.scene.objects + 1] = screen_group_obj

    -- Position and name of the screen
    result.scene.objects[#result.scene.objects + 1] = {
        name = consts.SCREEN,
        position = { 0, 0, 0 },
        orientation = quat.front_up_to_quaternion({ 0, 0, -1 }, { 0, 1, 0 }),
        openglmodel = "current.opengl.screen",
        relativeto = consts.SCREENGROUP,
        action = {
            trigger = {
                native = actions_consts.ACTION_ALIGN_AND_PAUSE, -- TODO/FIXME: not DRY!!!
            },
        }
    }
    local screen_points = {
        { -screen_params.width * 0.5, 0.5 * screen_params.height, 0 },
        { screen_params.width * 0.5, -0.5 * screen_params.height, 0 }
    }
    -- TODO/FIXME: eugh…
    -- save the index to inject some code as soon as we have the correct objects structure
    local screen_index = #result.scene.objects

    local camera_obj = {
        name = consts.CAMERA,
        position = { 0, screen_params.lift, -screen_params.distance },
        orientation = quat.front_up_to_quaternion({ 0, 0, -1 }, { 0, 1, 0 }),
        relativeto = consts.ROOM
    }

    local skybox_scene = create_skybox(camera_obj)
    -- cubemap walls
    result.scene.objects = functional.append_array(result.scene.objects,
        { skybox_scene })

    -- "resume" button on the screen. Will ignore the depth test
    result.scene.objects[#result.scene.objects + 1] = {
        name = consts.RESUME,
        visible = false,
        openglmodel = "current.opengl.resume",
        action = {
            trigger = {
                native = actions_consts.ACTION_RESUME,
            },
            vselected = {
                native = actions_consts.ACTION_RESUME,
            },
        },
        relativeto = consts.SCREEN,
        position = { 0, 0, 0.5 },
        orientation = quat.front_up_to_quaternion({ 0, 0, -1 }, { 0, 1, 0 }),
    }
    
    result.scene.objects[#result.scene.objects + 1] = scene_controls.create_seek_bar_cubemap(
       screen_params,
       vr_config.get(options_consts.SEEK_BAR_SHOW_BAR),
       vr_config.get(options_consts.SEEK_BAR_POSITION))

    result.scene.objects[#result.scene.objects + 1] = camera_obj

    result.scene.objects[#result.scene.objects + 1] = {
        name = consts.SUBTITLES_BASE,
        position = subtitles.compute_position(vr_config.get(options_consts.SUBTITLES_QUAD_DISTANCE),
            screen_points,
            camera_obj),
        orientation = quat.front_up_to_quaternion({ 0, 0, -1 }, { 0, 1, 0 }),
        relativeto = consts.ROOM,
    }
    local subs_base = result.scene.objects[#result.scene.objects]

    result.scene.objects[#result.scene.objects + 1] = {
        name = consts.SUBTITLES,
        visible = parameters.show_subtitles,
        openglmodel = "current.opengl.subtitles_quad",
        relativeto = consts.SUBTITLES_BASE,
        position = { 0, 0, 0 },
        orientation = quat.front_up_to_quaternion({ 0, 0, -1 }, { 0, 1, 0 }),
    }

    local function movement_function(displacement)
       -- TODO/FIXME: tutto da rifare
       -- TODO/FIXME: hard coded. Either remove the big cubemap size or do some kind of check!
       local LIMIT = cubemap_common.SMALL_CUBEMAP_SIZE
       local new_pos = mymath.add_3D(camera_obj.position, mymath.smul_3D(displacement, 10.0))
       if (math.abs(new_pos[2]) < LIMIT * 0.5) and (new_pos[3] > -(LIMIT * 0.99) and new_pos[3] < -10) then
          camera_obj.position = new_pos
          camera_obj.dirty = true
          skybox_scene.position = mymath.neg_3D(camera_obj.position)
          skybox_scene.dirty = true
          subs_base.position = subtitles.compute_position(
             vr_config.get(options_consts.SUBTITLES_QUAD_DISTANCE),
             screen_points,
             camera_obj)
          subs_base.dirty = true
          return true
       else
          return false
       end
    end

    local function save_config_function()
       vr_config.set(options_consts.CUBEMAP_SCREEN_LIFT, math.floor(camera_obj.position[2]))
       vr_config.set(options_consts.CUBEMAP_SCREEN_DISTANCE, math.floor(-camera_obj.position[3]))
    end

    local pitch = mymath.grad2rad(vr_config.get(options_consts.CONTROLS_PITCH))
    local yaw = mymath.grad2rad(vr_config.get(options_consts.CONTROLS_YAW))

    local extra_controls = scene_controls.create_theater_controls(
       camera_obj, options.ui, options.ui.buttons_radius,
       pitch, yaw,
       {
          movement = movement_function, 
          zooming = nil, 
          config_save = save_config_function
    })
    
    -- Controls
    result.scene.objects = functional.append_array(result.scene.objects, extra_controls)

    result.scene.objects[#result.scene.objects + 1] = {
        name = consts.PROGRESS,
        visible = false,
        openglmodel = "current.opengl.progress",
        position = { 0.0, 0.0, -6.0 },
        orientation = quat.front_up_to_quaternion({ 0, 0, -1}, { 0, 1, 0 })
    }

    -- TODO/FIXME: this sucks sooo big time :(
    result.scene.objects[screen_index].action.start_focus = scene_controls.generate_screen_start_focus(result.scene.objects)
    result.scene.objects[screen_index].action.end_focus = scene_controls.generate_screen_end_focus(result.scene.objects)

    -- Keep this last!
    result.scene.generic_callbacks = scene_controls.generate_generic_callbacks(result.scene.objects)

    result.opengl = rendering.create_models(parameters, screen_params, screen_points)

    return result
end

return {
    create = create
}
