local opengl       = require "opengl"
local options      = require "options"
local consts       = require "scene/scene_constants"
local scene_debug  = require "scene/scene_debug"
local vr_config    = require "scene/vr_config"
local common_uniforms = require "scene/common_uniforms"

local function direct_indirect_prefix(direct_rendering) 
   if direct_rendering then
      return "direct"
   else
      return "prerendered"
   end         
end


--   TODO/FIXME: DRY violation with Config.java (where the mode strings are defined!)
--   Returns a "shaders/common.vert" file fit for the conversion of the uv coordinates.
--
--   @param mode  the type of stereographic movie selected. Accepted modes are: "mono", "SBS", "TOP BOTTOM" etc
--   @returns  the name of the file to use in the vertex shader of the screen that contains the uv conversion function
local function screenCoordsShader(mode, direct_rendering)
   local direct_indirect = direct_indirect_prefix(direct_rendering)
   if mode == consts.ARG_STEREO_TYPE_MONO 
   or mode == consts.ARG_STEREO_TYPE_ANAGLYPH_RED_CYAN then
      return "shaders/screen/" .. direct_indirect .. "/mono_conversion.vert"
   elseif mode == consts.ARG_STEREO_TYPE_SBS then
      return "shaders/screen/" .. direct_indirect .. "/sbs_conversion.vert"
   elseif mode == consts.ARG_STEREO_TYPE_TOPBOTTOM then
      return "shaders/screen/" .. direct_indirect .. "/tb_conversion.vert"
   end
end

local function shader_suffix_for_aspect_ratio(ratio) 
   if ratio == 360.0 / 180.0 or ratio == 180.0 / 180.0 then
      return "panoramic"
   else 
      return "screen"
   end
end

local function shader_prefix_for_3D_mode(mode) 
   if mode == consts.ARG_STEREO_TYPE_ANAGLYPH_RED_CYAN then
      return "anaglyph"
   else
      return "plain"
   end
end

local function shader_name(mode, ratio, extension, direct_rendering)
   local direct_indirect = direct_indirect_prefix(direct_rendering)
   return "shaders/screen/" .. direct_indirect .. "/" .. shader_prefix_for_3D_mode(mode) .. "_" .. shader_suffix_for_aspect_ratio(ratio) .. "." .. extension
end

local function get_screen_vertex_shader(mode, ratio, direct_rendering)
   local direct_indirect = direct_indirect_prefix(direct_rendering)
   return {
      "shaders/screen/" .. direct_indirect .. "/common_screen.vert", 
      screenCoordsShader(mode, direct_rendering), 
      shader_name(mode, ratio, "vert", direct_rendering)
   }
end

local function get_screen_fragment_shader(mode, ratio, direct_rendering)
   return {
      shader_name(mode, ratio, "frag", direct_rendering)
   }
end

local function anaglyph_conversion(color)
   if color == consts.COLOR_RED then
      return {
	 1.0, 0.0, 0.0,
	 0.0, 0.0, 0.0,
	 0.0, 0.0, 0.0
      }
   end 
   if color == consts.COLOR_CYAN then
      return {
	 0.0, 0.0, 0.0,
	 0.0, 1.0, 0.0,
	 0.0, 0.0, 1.0
      }
   end

   -- Hopefully, this will scream "error" when seen
   return { 
      0.3, 0.7, 0.0,
      0.6, 0.2, 0.2,
      0.2, 0.3, 0.5
   }
end

local function progress_bar_model(uniforms) 
   local progress_width = 0.4
   local progress_height = 0.05
   return {
      {
         type = "bind program",
         shader = {
            vert = "shaders/progress/progress.vert", 
            frag = "shaders/progress/progress.frag",
         }
      },
      { -- uses the current shader
         type = "bind uniforms",
         
         uniforms = {
            model = common_uniforms.model,
            projection = common_uniforms.projection,
            activation  = uniforms.activation
         }
      },
      {
         type = "bind attribute",
         name = "pos",
         values = {
            (-progress_width * 0.5), -0.3 + progress_height * 0.5,    0, 
            (-progress_width * 0.5), -0.3 + (-progress_height * 0.5), 0,
            progress_width * 0.5,    -0.3 + (-progress_height * 0.5), 0,
            progress_width * 0.5,    -0.3 + progress_height * 0.5,    0},
         dim = 3
      },
      {
         type = "bind attribute",
         name = "uv",
         values = {
            0, 0, 1, 1
         },
         dim = 1
      },
      {
         type = "disable",
         name = "GL_DEPTH_TEST"
      },
      {
         type = "draw",     
         primitive = "GL_TRIANGLES",
         elements = {0, 1, 2, 0, 2, 3},
      },
      {
         type = "enable",
         name = "GL_DEPTH_TEST"
      },
      {  
         type = "bind attribute",
         name = "uv"
      },
      {  
         type = "bind attribute",
         name = "pos"
      }
   }
end

return {
   anaglyph_conversion = anaglyph_conversion,
   progress_bar_model = progress_bar_model,
   get_screen_vertex_shader = get_screen_vertex_shader,
   get_screen_fragment_shader = get_screen_fragment_shader,
}
