-- Scene configuration for panoramic modes
-- Created by Riccardo Di Meo
-- © Riccardo Di Meo 2015
local actions_consts = require "scene/actions_consts"
local consts = require "scene/scene_constants"
local functional = require "scene/functional"
local mymath = require "mymath"
local opengl = require "opengl"
local options = require "options"
local options_consts = require "scene/options_consts"
local quat = require "scene/quaternion"
local rendering = require "scene/panoramic/rendering"
local scene_controls = require "scene/controls/controls"
local vr_config = require "scene/vr_config"

local SPHERE_SIZE = 200

local function create(parameters)
    opengl.znear = 0.1
    opengl.zfar = 300.0


    -- should the tigger/tap align the screen in panoramic mode?
    local trigger_response
    if vr_config.get(options_consts.PANORAMIC_MODE_ALIGNMENT) then
        trigger_response = actions_consts.ACTION_ALIGN_AND_PAUSE
    else
        trigger_response = actions_consts.ACTION_PAUSE
    end

    local current = {}
    current.scene = {}

    current.scene.player = consts.CAMERA
    current.scene.locked = consts.SCREEN -- not actually used

    current.scene.objects = {}

    -- Position and name of the screen
    local screen_obj = {
        name = consts.SCREEN,
        position = { 0, 0, 0 }, -- Note; position is not intended in the usual sense: it's the zoom (shader details)
        scale = { SPHERE_SIZE, SPHERE_SIZE, SPHERE_SIZE },
        orientation = quat.front_up_to_quaternion({ 0, 0, -1 }, { 0, 1, 0 }),
        openglmodel = "current.opengl.screen",
        action = {
            trigger = {
                native = trigger_response,
            },
        }
    }
    current.scene.objects[#current.scene.objects + 1] = screen_obj

    -- TOFO/FIXME: eugh…
    -- save the index to inject some code as soon as we have the correct objects structure
    local screen_index = #current.scene.objects

    -- Position of the observer
    current.scene.objects[#current.scene.objects + 1] = {
        name = consts.CAMERA,
        position = { 0, 0, 0 },
        orientation = quat.front_up_to_quaternion({ 0, 0, -1 }, { 0, 1, 0 }),
    }
    local camera_obj = current.scene.objects[#current.scene.objects]

    current.scene.objects[#current.scene.objects + 1] = {
        name = consts.RESUME,
        visible = false,
        openglmodel = "current.opengl.resume",
        action = {
            trigger = {
                native = actions_consts.ACTION_RESUME,
            },
            vselected = {
                native = actions_consts.ACTION_RESUME,
            },
        },
        position = { 0.0, 0.0, -15.0 },
        orientation = quat.front_up_to_quaternion({ 0, 0, -1 }, { 0, 1, 0 }),
    }

    local zoom_function = function(zoom_factor) -- zoom_factor == nil is a special value that stands for "reset"
       if zoom_factor == nil then
          screen_obj.position[3] = 0.0
          screen_obj.dirty = true
          return true
       else
          local next_value = screen_obj.position[3] + zoom_factor
          if next_value < 150 and next_value > -90 then
             screen_obj.position[3] = next_value
             print("Zoom function with factor: " .. tostring(next_value))
             screen_obj.dirty = true
             return true
          else
             return false
          end
       end
    end

    -- TODO replace this with a nil!
    local save_config_function = function() end

    local pitch = mymath.grad2rad(vr_config.get(options_consts.PANORAMIC_CONTROLS_PITCH))
    local yaw = mymath.grad2rad(vr_config.get(options_consts.PANORAMIC_CONTROLS_YAW))

    local extra_controls = scene_controls.create_panoramic_controls(
       camera_obj, options.ui, options.ui.buttons_radius,
       pitch, yaw,
       {
          movement = nil, 
          zooming = zoom_function, 
          config_save = save_config_function
    })

    local function remove_lock_theater_entity(controls)
       return functional.filter(
          controls,
          function(entity)
             return entity.name ~= consts.LOCK_THEATER
       end)
    end

    current.scene.objects = functional.append_array(
       current.scene.objects,
       remove_lock_theater_entity(extra_controls))

    -- Activation feedback
    current.scene.objects[#current.scene.objects + 1] = {
        name = consts.PROGRESS,
        visible = false,
        openglmodel = "current.opengl.progress",
        scale = {60, 60, 1},
        position = { 0.0, 0.0, -150.0 },
        orientation = quat.front_up_to_quaternion({ 0, 0, -1 }, { 0, 1, 0 })
    }

    -- TODO/FIXME: this sucks sooo big time :(
    current.scene.objects[screen_index].action.start_focus = scene_controls.generate_screen_start_focus(current.scene.objects)
    current.scene.objects[screen_index].action.end_focus = scene_controls.generate_screen_end_focus(current.scene.objects)


    current.scene.generic_callbacks = scene_controls.generate_generic_callbacks(current.scene.objects)

    current.opengl = rendering.create_opengl_objects(parameters)

    return current
end

return {
    create = create
}
