local consts = require "scene/scene_constants"
local vr_config = require "scene/vr_config"
local options = require "options"
local options_consts = require "scene/options_consts"
local opengl = require "opengl" -- TODO/FIXME: fixme!
local common_uniforms = require "scene/common_uniforms"

local function float_equal(a, b)
    return math.abs(a - b) < 0.0001
end

local function return_pointer(side, opacity)

    local size = cardboardtheater.get_screen_size()
    local ratio = 0.5 * size.height / size.width

    local width = side * size.height / size.width
    local height = side * 0.5

    return {
        {
            type = "bind program",
            shader = {
                vert = "shaders/pointer/pointer.vert",
                frag = "shaders/pointer/pointer.frag",
            }
        },
        {
            -- uses the current shader
            type = "bind uniforms",
            uniforms = {
                image = {
                    content = "TEX:1:" .. opengl.texturedir .. "/pointer/pointer.png",
                    vtype = "int"
                },
                model = common_uniforms.model,
                view = common_uniforms.view,
                projection = common_uniforms.projection,
                alpha = {
                    content = opacity,
                    vtype = "float"
                }
            }
        },
        {
            type = "bind attribute",
            name = "pos",
            values = {
                (-width), (-height), 0,
                (-width), (height), 0,
                (width), (height), 0,
                (width), (-height), 0
            },
            dim = 3
        },
        {
            type = "bind attribute",
            name = "uv",
            values = {
                0, 0,
                0, 1,
                1, 1,
                1, 0
            },
            dim = 2
        },
--        {
--            type = "disable",
--            name = "GL_DEPTH_TEST"
--        },
        {
            type = "draw",
            primitive = "GL_TRIANGLES",
            elements = { 0, 1, 2, 2, 3, 0 },
        },
--      {
--          type = "enable",
--          name = "GL_DEPTH_TEST"
--      },
        {
            type = "bind attribute",
            name = "uv"
        },
        {
            type = "bind attribute",
            name = "pos"
        }
    }
end

-- I use a generator function to isolate the buffers and buffering function from the rest of the code
local return_active_object = (function()
    local active_objects_index = { 0, 1, 2, 2, 3, 0 }
    local active_objects_uv = {
        values = {
            0, 0,
            0, 1,
            1, 1,
            1, 0
        },
        dim = 2
    }
    local saved_pos_attributes = {}

    local function get_attribute_with_dimensions(width, height)
        local key = width .. "x" .. height
        local cached = saved_pos_attributes[key]
        if cached == nil then
            local buffer = {
                values = {
                    -width, -height, 0,
                    -width, height, 0,
                    width, height, 0,
                    width, -height, 0
                },
                dim = 3
            }
            saved_pos_attributes[key] = buffer
            return buffer
        else
            return cached
        end
    end

    local function bind_shader(icons)
       if icons.over then
          return {
             type = "bind program",
             shader = {
                vert = "shaders/active_two_states_button/active_two_states_button.vert",
                frag = "shaders/active_two_states_button/active_two_states_button.frag",
             }
          }
       else
          return {
             type = "bind program",
             shader = {
                vert = "shaders/activesquare/activesquare.vert",
                frag = "shaders/activesquare/activesquare.frag",
             }
          }
       end
    end

    local function bind_uniforms(width, height, icons)
       local controls_opacity = { content = 1.0, vtype = "float" }
       local activated = { content = "VAR:ACTIVATED", vtype = "int" }
       local controls_modelview = { content = "VAR:CONTROLS_MODELVIEW", vtype = "mat4" }

       if icons.over then
          return {
             -- uses the current shader
             type = "bind uniforms",
             uniforms = {
                normal_icon = icons.image,
                over_icon = icons.over,
                controls_opacity = controls_opacity,
                activated = activated,
                controls_modelview = controls_modelview,
                projection = common_uniforms.projection,
                id = common_uniforms.id,
                feedback_rendering = common_uniforms.feedback_rendering,
             }
          }
       else
          return {
             -- uses the current shader
             type = "bind uniforms",
             uniforms = {
                icon = icons.image,
                controls_opacity = controls_opacity,
                activated = activated,
                controls_modelview = controls_modelview,
                projection = common_uniforms.projection,
                id = common_uniforms.id,
                feedback_rendering = common_uniforms.feedback_rendering,
             }
          }          
       end
    end

    return function(width, height, icons)
       local pos = get_attribute_with_dimensions(width, height)
       
       return {
          bind_shader(icons),
          bind_uniforms(width, height, icons),
          {
             type = "bind attribute",
             name = "pos",
             values = pos.values,
             dim = pos.dim
          },
          {
             type = "bind attribute",
             name = "uv",
             values = active_objects_uv.values,
             dim = active_objects_uv.dim
          },
          {
             type = "disable",
             name = "GL_DEPTH_TEST"
          },
          {
             type = "draw",
             primitive = "GL_TRIANGLES",
             elements = active_objects_index,
          },
          {
             type = "enable",
             name = "GL_DEPTH_TEST"
          },
          {
             type = "bind attribute",
             name = "uv"
          },
          {
             type = "bind attribute",
             name = "pos"
          }
       }
    end
end)()

local function _generate_panel_glmodel(square_side, color, texture)
    return {
        {
            type = "bind program",
            shader = {
                vert = "shaders/panel/panel.vert",
                frag = "shaders/panel/panel.frag"
            }
        },
        {
            -- uses the current shader
            type = "bind uniforms",
            uniforms = {
                id = common_uniforms.id,
                modelview = {
                    content = "VAR:CONTROLS_MODELVIEW",
                    vtype = "mat4",
                },
                opacity = {
                    content = (vr_config.get(options_consts.ACTIVE_AREAS_OPACITY) / 100),
                    vtype = "float"
                },
                icon = {
                    content = "TEX:1:" .. opengl.texturedir .. "/active_areas/" .. texture,
                    vtype = "int"
                },
                color = {
                    content = color or { 1.0, 1.0, 1.0, 0.8 },
                    vtype = "vec4"
                },
                projection = common_uniforms.projection,
                feedback_rendering = common_uniforms.feedback_rendering,
            }
        },
        {
            type = "bind attribute",
            name = "pos",
            values = {
                (-square_side * 0.5), 0.5 * square_side, 0,
                (-square_side * 0.5), -0.5 * square_side, 0,
                square_side * 0.5, -0.5 * square_side, 0,
                square_side * 0.5, 0.5 * square_side, 0,
            },
            dim = 3
        },
        {
            type = "bind attribute",
            name = "uv",
            values = {
                0, 0,
                0, 1,
                1, 1,
                1, 0
            },
            dim = 2
        },
        {
            type = "disable",
            name = "GL_DEPTH_TEST"
        },
        {
            type = "draw",
            primitive = "GL_TRIANGLES",
            elements = { 0, 1, 2, 0, 2, 3 },
        },
        {
            type = "enable",
            name = "GL_DEPTH_TEST"
        },
        {
            type = "bind attribute",
            name = "uv"
        },
        {
            type = "bind attribute",
            name = "pos"
        }
    }
end

local function generate_panel_space_glmodel(side, color, texture)
    return _generate_panel_glmodel(side, color, "drop.png")
end

local function generate_panel_button_glmodel(side, color, texture)
    return _generate_panel_glmodel(side, color, "circle.png")
end

-- Not DRY with MovieSwitcher
local function create_screen_type_opengl_objects(collection, current_ratio)
    local current_screen_type_texture, next_screen_type_texture
    if current_ratio == 2 or current_ratio == 1 then
        current_screen_type_texture = "TEX:1:" .. opengl.texturedir .. "/active_areas/pano_screen.png"
        next_screen_type_texture = "TEX:2:" .. opengl.texturedir .. "/active_areas/theater_screen.png"
    else
        current_screen_type_texture = "TEX:1:" .. opengl.texturedir .. "/active_areas/theater_screen.png"
        next_screen_type_texture = "TEX:2:" .. opengl.texturedir .. "/active_areas/pano_screen.png"
    end
    local icons = {
       image = {content = current_screen_type_texture, vtype = "int" },
       over = {content = next_screen_type_texture, vtype = "int" }
    }
    collection.screen_type = return_active_object(options.ui.button_side, options.ui.button_side, icons)
end

-- Not DRY with MovieSwitcher
local function create_screen_ratio_opengl_objects(collection, current_ratio)
   local icons
   if current_ratio == 1 then
      icons = {
         image = { content = "TEX:1:" .. opengl.texturedir .. "/active_areas/180.png", vtype = "int"},
         over = { content = "TEX:2:" .. opengl.texturedir .. "/active_areas/360.png", vtype = "int"}
      }
   elseif current_ratio == 2 then
      icons = {
         image = { content = "TEX:1:" .. opengl.texturedir .. "/active_areas/360.png", vtype = "int"},
         over = { content = "TEX:2:" .. opengl.texturedir .. "/active_areas/180.png", vtype = "int"}
      }
   elseif float_equal(current_ratio, 9 / 16) then      
      icons = {
         image =  { content = "TEX:1:" .. opengl.texturedir .. "/active_areas/9_16.png", vtype = "int"},
         over =  { content = "TEX:2:" .. opengl.texturedir .. "/active_areas/4_3.png", vtype = "int"}
      }
   elseif float_equal(current_ratio, 4 / 3) then      
      icons = {
         image =  { content = "TEX:1:" .. opengl.texturedir .. "/active_areas/4_3.png", vtype = "int"},
         over =  { content = "TEX:2:" .. opengl.texturedir .. "/active_areas/16_10.png", vtype = "int"}
      }
   elseif float_equal(current_ratio, 16 / 10) then
      icons = {
         image =  { content = "TEX:1:" .. opengl.texturedir .. "/active_areas/16_10.png", vtype = "int"},
         over =  { content = "TEX:2:" .. opengl.texturedir .. "/active_areas/16_9.png", vtype = "int"}
      }
   elseif float_equal(current_ratio, 16 / 9) then
      icons = {
         image =  { content = "TEX:1:" .. opengl.texturedir .. "/active_areas/16_9.png", vtype = "int"},
         over =  { content = "TEX:2:" .. opengl.texturedir .. "/active_areas/239_1.png", vtype = "int"}
      }
   else
      icons = {
         image =  { content = "TEX:1:" .. opengl.texturedir .. "/active_areas/239_1.png", vtype = "int"},
         over =  { content = "TEX:2:" .. opengl.texturedir .. "/active_areas/9_16.png", vtype = "int"}
      }
   end

   collection.screen_ratio = return_active_object(options.ui.button_side, options.ui.button_side, icons)
end


-- Not DRY with MovieSwitcher
local function create_movie_3D_conversion_opengl_objects_from_current_3D(collection, current_3D)
    local next_3D_type_texture, current_3D_type_texture

    if current_3D == consts.ARG_STEREO_TYPE_MONO then
        current_3D_type_texture = "TEX:1:" .. opengl.texturedir .. "/active_areas/2d.png"
        next_3D_type_texture = "TEX:2:" .. opengl.texturedir .. "/active_areas/sbs.png"
    elseif current_3D == consts.ARG_STEREO_TYPE_SBS then
        current_3D_type_texture = "TEX:1:" .. opengl.texturedir .. "/active_areas/sbs.png"
        next_3D_type_texture = "TEX:2:" .. opengl.texturedir .. "/active_areas/tab.png"
    elseif current_3D == consts.ARG_STEREO_TYPE_TOPBOTTOM then
        current_3D_type_texture = "TEX:1:" .. opengl.texturedir .. "/active_areas/tab.png"
        next_3D_type_texture = "TEX:2:" .. opengl.texturedir .. "/active_areas/red_cyan.png"
    elseif current_3D == consts.ARG_STEREO_TYPE_ANAGLYPH_RED_CYAN then
        current_3D_type_texture = "TEX:1:" .. opengl.texturedir .. "/active_areas/red_cyan.png"
        next_3D_type_texture = "TEX:2:" .. opengl.texturedir .. "/active_areas/2d.png"
    else
        current_3D_type_texture = "TEX:1:" .. opengl.texturedir .. "/active_areas/2d.png"
        next_3D_type_texture = "TEX:2:" .. opengl.texturedir .. "/active_areas/2d.png"
    end

    local icons = {
       image = {content = current_3D_type_texture, vtype = "int" },
       over  = {content = next_3D_type_texture, vtype = "int" }
    }

    collection.stereo_type = return_active_object(options.ui.button_side, options.ui.button_side, icons)
end

local function generate_control_objects(resume_size_or_nil, movement, zoom, current_ratio, current_3D)
    local result = {}

    -- Create the controls
    result.invert = return_active_object(options.ui.button_side, options.ui.button_side, {image = {content = "TEX:1:" .. opengl.texturedir .. "/active_areas/invert.png", vtype = "int" }})
    result.forward = return_active_object(options.ui.button_side, options.ui.button_side, {image = {content = "TEX:1:" .. opengl.texturedir .. "/active_areas/skip_forward.png", vtype = "int" }})
    result.backward = return_active_object(options.ui.button_side, options.ui.button_side, {image = {content = "TEX:1:" .. opengl.texturedir .. "/active_areas/skip_backward.png", vtype = "int" }})
    result.fast_forward = return_active_object(options.ui.button_side, options.ui.button_side, {image = {content = "TEX:1:" .. opengl.texturedir .. "/active_areas/fast_forward.png", vtype = "int" }})
    result.fast_backward = return_active_object(options.ui.button_side, options.ui.button_side, {image = {content = "TEX:1:" .. opengl.texturedir .. "/active_areas/fast_backward.png", vtype = "int" }})
    result.align = return_active_object(options.ui.button_side, options.ui.button_side, {image = {content = "TEX:1:" .. opengl.texturedir .. "/active_areas/align.png", vtype = "int" }})
    result.volumeup = return_active_object(options.ui.button_side, options.ui.button_side / 2, {image = {content = "TEX:1:" .. opengl.texturedir .. "/active_areas/vup.png", vtype = "int" }})
    result.volumedown = return_active_object(options.ui.button_side, options.ui.button_side / 2, {image = {content = "TEX:1:" .. opengl.texturedir .. "/active_areas/vdown.png", vtype = "int" }})
    result.lock = return_active_object(options.ui.button_side, options.ui.button_side, {image = {content = "TEX:1:" .. opengl.texturedir .. "/active_areas/lock.png", vtype = "int" }})
    result.pause = return_active_object(options.ui.button_side, options.ui.button_side, {image = {content = "TEX:1:" .. opengl.texturedir .. "/active_areas/pause.png", vtype = "int" }})
    -- second group

    result.prev_track = return_active_object(options.ui.button_side, options.ui.button_side, {image = {content = "TEX:1:" .. opengl.texturedir .. "/active_areas/prev_track.png", vtype = "int" }})

    result.next_track = return_active_object(options.ui.button_side, options.ui.button_side, {image = {content = "TEX:1:" .. opengl.texturedir .. "/active_areas/next_track.png", vtype = "int" }})

    -- zoom and movement are conditional to the kind of scene shown
    if movement then
        result.move_forward = return_active_object(options.ui.button_side, options.ui.button_side / 2, {image = {content = "TEX:1:" .. opengl.texturedir .. "/active_areas/forward.png", vtype = "int" }})
        result.move_backward = return_active_object(options.ui.button_side, options.ui.button_side / 2, {image = {content = "TEX:1:" .. opengl.texturedir .. "/active_areas/backward.png", vtype = "int" }})
        result.move_up = return_active_object(options.ui.button_side, options.ui.button_side / 2, {image = {content = "TEX:1:" .. opengl.texturedir .. "/active_areas/up.png", vtype = "int" }})
        result.move_down = return_active_object(options.ui.button_side, options.ui.button_side / 2, {image = {content = "TEX:1:" .. opengl.texturedir .. "/active_areas/down.png", vtype = "int" }})
    elseif zoom then
        result.zoom_in = return_active_object(options.ui.button_side, options.ui.button_side / 2, {image = {content = "TEX:1:" .. opengl.texturedir .. "/active_areas/zoom_in.png", vtype = "int" }})
        result.zoom_out = return_active_object(options.ui.button_side, options.ui.button_side / 2, {image = {content = "TEX:1:" .. opengl.texturedir .. "/active_areas/zoom_out.png", vtype = "int" }})
        result.reset_zoom = return_active_object(options.ui.button_side, options.ui.button_side, {image = {content = "TEX:1:" .. opengl.texturedir .. "/active_areas/reset_zoom.png", vtype = "int" }})
    end

    -- screen type configuration
    create_screen_type_opengl_objects(result, current_ratio)
    -- screen ratio configuration
    create_screen_ratio_opengl_objects(result, current_ratio)
    -- movie 3D conversion
    create_movie_3D_conversion_opengl_objects_from_current_3D(result, current_3D)


    -- plus button
    result.switch_button = return_active_object(options.ui.button_side, options.ui.button_side, {image = {content = "TEX:1:" .. opengl.texturedir .. "/active_areas/more.png", vtype = "int" }})

    -- this one is a bit different…
    local resume_size
    if resume_size_or_nil then
        resume_size = resume_size_or_nil
    else
        resume_size = 4.0
    end
    result.resume = return_active_object(resume_size, resume_size, {image = {content = "TEX:1:" .. opengl.texturedir .. "/active_areas/play.png", vtype = "int" }})

    -- TODO/FIXME: the 0.5 extra factor near the button_size (it should be 2) should go in the options
    result.controls_panel = generate_panel_space_glmodel(options.ui.buttons_radius * 2.5 + 2 * options.ui.button_side)
    result.controls_button = generate_panel_button_glmodel(vr_config.get(options_consts.CONTROLS_TRIGGER_SIZE) / 100.0, { 1.0, 0.0, 0.0, 0.9 })

    return result
end

return {
    generate_control_objects = generate_control_objects,
    return_pointer = return_pointer
}
