local opengl = require "opengl"
local common_uniforms = require "scene/common_uniforms"
local consts = require "scene/scene_constants"
local functional = require "scene/functional"
local subtitles = require "scene/subtitles"
local vr_config = require "scene/vr_config"
local options_consts = require "scene/options_consts"
local scene_common = require "scene/scene_common"

local opengl_common = require "scene/opengl_common" --- !!!!! TODO/FIXME!!!!!
local scene_controls_opengl = require "scene/scene_controls_opengl" --- !!! TODO/FIXME !!!!

local function create_opengl_objects(parameters, screen_points, screen_params)

    local scene_uniforms
    if vr_config.get(options_consts.THEATER_TYPE) == options_consts.THEATER_ENV then
        scene_uniforms = scene_common.return_cardboard_theater_environment_uniforms(screen_params)
    else
        scene_uniforms = scene_common.return_cardboard_cardboard_theater_environment_uniforms(screen_params)
    end

    opengl.znear = 0.1
    opengl.zfar = 300.0

    local stereo_type = parameters.stereo_type
    local direct_rendering = parameters.content_type ~= "image"

    local texsurface_common = function(prefix, vertshader)
        return {
            programs = {
                vert = vertshader,
                frag = "shaders/texsurface/texsurface.frag",
            },
            uniforms = {
                projection = common_uniforms.projection,
                modelview = common_uniforms.modelview,
                model = common_uniforms.model,
                view = common_uniforms.view,
                time = common_uniforms.time,

                uvscale = scene_uniforms[prefix .. "_uvscale"],
                uvoffset = scene_uniforms[prefix .. "_uvoffset"],
                diffuse = scene_uniforms[prefix .. "_diffuse"],
                normal = scene_uniforms[prefix .. "_normal"],
                specular = scene_uniforms[prefix .. "_specular"],
                --	    bump = scene_uniforms[prefix .. "_bump"],
                phongweights = scene_uniforms[prefix .. "_phongweights"] or scene_uniforms["common_phongweights"],
                ambientcolor = scene_uniforms[prefix .. "_ambientcolor"] or scene_uniforms["common_ambientcolor"],
                lightpos = scene_uniforms[prefix .. "_lightpos"] or scene_uniforms["common_lightpos"],
                lightanimpar = scene_uniforms[prefix .. "_lightanimpar"] or scene_uniforms["common_lightanimpar"],
                lightparams = scene_uniforms[prefix .. "_lightparams"] or scene_uniforms["common_lightparams"],
                lightrange = scene_uniforms[prefix .. "_lightrange"],
            }
        }
    end

    local objects = {}

    -- This will set debug_objects to a couple of useful debug OpenGL Objects to be used with debugmodel
    objects.debug_objects = opengl.generate_debug_objects(common_uniforms)

    objects.opengl = {
        axes = {
            {
                type = "bind program",
                shader = {
                    vert = "shaders/axes/axes.vert",
                    frag = "shaders/axes/axes.frag",
                }
            },
            {
                type = "bind uniforms",
                uniforms = {
                    projection = common_uniforms.projection,
                    modelview = common_uniforms.modelview
                }
            },
            {
                type = "bind attribute",
                name = "pos",
                values = {
                    0, 0, 0,
                    1, 0, 0,
                    0.9, 0.1, 0,
                    0.9, 0, 0.1,
                    0, 1, 0,
                    0.1, 0.9, 0,
                    0, 0.9, 0.1,
                    0, 0, 1,
                    0.1, 0, 0.9,
                    0, 0.1, 0.9
                },
                dim = 3
            },
            {
                type = "bind attribute",
                name = "vcolor",
                values = {
                    1, 1, 1,
                    1, 0, 0,
                    1, 0, 0,
                    1, 0, 0,
                    0, 1, 0,
                    0, 1, 0,
                    0, 1, 0,
                    0, 0, 1,
                    0, 0, 1,
                    0, 0, 1,
                },
                dim = 3
            },
            {
                type = "draw",
                primitive = "GL_LINES",
                elements = {
                    0, 1, 1, 2, 1, 3,
                    0, 4, 4, 5, 4, 6,
                    0, 7, 7, 8, 7, 9
                },
            }
        },
        -- Don't remove: this dot shows an example on how
        -- to place a dot a the center of the player's view
        dot = {
            {
                type = "bind program",
                shader = {
                    vert = "shaders/point/point.vert",
                    frag = "shaders/point/point.frag",
                }
            },
            {
                -- uses the current shader
                type = "bind uniforms",
                uniforms = {
                    projection = common_uniforms.projection
                }
            },
            {
                type = "bind attribute",
                name = "vcolor",
                values = {
                    1.0, 0, 0,
                },
                dim = 3
            },
            {
                type = "draw",
                primitive = "GL_POINTS",
                count = 1,
            },
            {
                type = "bind attribute",
                name = "vcolor"
            },
        }
    }

    local create_screen = function()
        local uniforms = {
            projection = common_uniforms.projection,
            modelview = common_uniforms.modelview,
            eye = common_uniforms.eye,
            id = common_uniforms.id,
            brightness = common_uniforms.brightness,
            contrast = common_uniforms.contrast,
            feedback_rendering = common_uniforms.feedback_rendering,
        }
        -- different shaders for different situations
        if direct_rendering then
            uniforms.video = common_uniforms.video
            uniforms.video_transform = common_uniforms.video_transform
        else
            uniforms.prevideo = common_uniforms.video
        end


        -- those two are used only when "anaglyph red/cyan" is selected
        if stereo_type == consts.ARG_STEREO_TYPE_ANAGLYPH_RED_CYAN then
            uniforms.swizzle_right = { content = opengl_common.anaglyph_conversion(consts.COLOR_RED), vtype = "mat3" }
            uniforms.swizzle_left = { content = opengl_common.anaglyph_conversion(consts.COLOR_CYAN), vtype = "mat3" }
        end

        local screen = {
            {
                type = "bind program",
                shader = {
                    vert = opengl_common.get_screen_vertex_shader(stereo_type, parameters.ratio, direct_rendering),
                    frag = opengl_common.get_screen_fragment_shader(stereo_type, parameters.ratio, direct_rendering)
                }
            },
            {
                -- uses the current shader
                type = "bind uniforms",
                uniforms = uniforms
            },
            {
                type = "bind attribute",
                name = "pos",
                values = {
                    (-0.5), 0.5, 0,
                    (-0.5), (-0.5), 0,
                    0.5, (-0.5), 0,
                    0.5, 0.5, 0
                },
                dim = 3
            },
            {
                type = "bind attribute",
                name = "uv",
                values = {
                    0, 1,
                    0, 0,
                    1, 0,
                    1, 1
                },
                dim = 2
            },
            {
                type = "draw",
                primitive = "GL_TRIANGLES",
                elements = { 0, 1, 2, 0, 2, 3 },
            },
            {
                type = "bind attribute",
                name = "uv"
            },
            {
                type = "bind attribute",
                name = "pos"
            }
        }
        return screen
    end
    objects.opengl.screen = create_screen()

    -- duplicated code with cubemap/renderer.lua
    function seek_bar_fragment_shader(seek_bar_position)
       if seek_bar_position == options_consts.SEEK_BAR_POSITION_TOP or seek_bar_position == options_consts.SEEK_BAR_POSITION_BOTTOM then
          return "shaders/seekbar/seekbar_horiz.frag"
       else
          return "shaders/seekbar/seekbar_vert.frag"
       end       
    end

    objects.opengl.seekbar =  {
       {
          type = "bind program",
          shader = {
             vert = "shaders/seekbar/seekbar.vert",
             frag = seek_bar_fragment_shader(vr_config.get(options_consts.SEEK_BAR_POSITION))
          }
       },
       {
          -- uses the current shader
          type = "bind uniforms",
          uniforms = {
             id = common_uniforms.id,
             movietime = common_uniforms.movietime,
             modelview = {
                content = "VAR:CONTROLS_MODELVIEW",
                vtype = "mat4",
             },
             opacity = {
                content = (vr_config.get(options_consts.SEEK_BAR_OPACITY) / 100),
                vtype = "float"
             },
             projection = common_uniforms.projection,
             feedback_rendering = common_uniforms.feedback_rendering,
          }
       },
       {
          type = "bind attribute",
          name = "pos",
          values = {
             (-0.5),  0.5, 0,
             (-0.5),  (-0.5), 0,
             0.5,     (-0.5), 0,
             0.5,     0.5, 0
          },
          dim = 3
       },
       {
          type = "bind attribute",
          name = "uv",
          values = {
             0, 1,
             0, 0,
             1, 0,
             1, 1
          },
          dim = 2
       },
       {
          type = "draw",
          primitive = "GL_TRIANGLES",
          elements = { 0, 1, 2, 0, 2, 3 },
       },
       {
          type = "bind attribute",
          name = "uv"
       },
       {
          type = "bind attribute",
          name = "pos"
       }
    }
    
    local floor_common_stuff = texsurface_common("floor", "shaders/texsurface/texsurfaceX.vert")
    objects.opengl.floor = {
        {
            type = "bind program",
            shader = floor_common_stuff.programs
        },
        {
            -- uses the current shader
            type = "bind uniforms", -- exception? Multiple uniforms, since the single uniform case is unlikely
            uniforms = floor_common_stuff.uniforms
        },
        {
            type = "bind attribute",
            name = "pos",
            values = {
                (-0.5), 0.5, 0,
                (-0.5), (-0.5), 0,
                0.5, (-0.5), 0,
                0.5, 0.5, 0
            },
            dim = 3
        },
        {
            type = "bind attribute",
            name = "uv",
            values = { 0, 1, 0, 0, 1, 0, 1, 1 },
            dim = 2
        },
        {
            -- good but not perfect, maybe. Should the index be put in a separate task?
            type = "draw",
            primitive = "GL_TRIANGLES", -- optional! GL_TRIANGLES is the default
            offset = 0, -- optional (default: 0)!
            --count  = 6, -- optional for indexed draws!
            elements = { 0, 1, 2, 2, 3, 0 }, -- optional! use draw elements otherwise
        },
        {
            -- no values -> unbind/disable attribute
            type = "bind attribute",
            name = "pos"
        },
        {
            -- no values -> unbind/disable attribute
            type = "bind attribute",
            name = "uv"
        }
    }
    floor_common_stuff = nil

    local rightwall_common_stuff = texsurface_common("rightwall", "shaders/texsurface/texsurfaceY.vert")
    objects.opengl.rightwall = {
        {
            type = "bind program",
            shader = rightwall_common_stuff.programs
        },
        {
            -- uses the current shader
            type = "bind uniforms", -- exception? Multiple uniforms, since the single uniform case is unlikely
            uniforms = rightwall_common_stuff.uniforms
        },
        {
            type = "bind attribute",
            name = "pos",
            values = {
                (-0.5), 0.5, 0.0,
                (-0.5), -(0.5), 0.0,
                0.5, -(0.5), 0.0,
                0.5, 0.5, 0.0,
            },
            dim = 3
        },
        {
            type = "bind attribute",
            name = "uv",
            values = { 0, 0, 0, 1, 1, 1, 1, 0 },
            dim = 2
        },
        {
            type = "draw",
            primitive = "GL_TRIANGLES",
            elements = { 0, 1, 2, 0, 2, 3 },
        },
        {
            -- no values -> unbind/disable attribute
            type = "bind attribute",
            name = "pos"
        },
        {
            -- no values -> unbind/disable attribute
            type = "bind attribute",
            name = "uv"
        }
    }
    rightwall_common_stuff = nil

    -- TODO/FIXME: almost identical to the rightwall, except the data in the texsurface_common (same attributes)
    local leftwall_common_stuff = texsurface_common("leftwall", "shaders/texsurface/texsurfaceY.vert")
    objects.opengl.leftwall = {
        {
            type = "bind program",
            shader = leftwall_common_stuff.programs
        },
        {
            -- uses the current shader
            type = "bind uniforms", -- exception? Multiple uniforms, since the single uniform case is unlikely
            uniforms = leftwall_common_stuff.uniforms
        },
        {
            type = "bind attribute",
            name = "pos",
            values = {
                (-0.5), 0.5, 0.0,
                (-0.5), (-0.5), 0.0,
                0.5, (-0.5), 0.0,
                0.5, 0.5, 0.0,
            },
            dim = 3
        },
        {
            type = "bind attribute",
            name = "uv",
            values = { 0, 0, 0, 1, 1, 1, 1, 0 },
            dim = 2
        },
        {
            type = "draw",
            primitive = "GL_TRIANGLES",
            elements = { 0, 1, 2, 0, 2, 3 },
        },
        {
            -- no values -> unbind/disable attribute
            type = "bind attribute",
            name = "pos"
        },
        {
            -- no values -> unbind/disable attribute
            type = "bind attribute",
            name = "uv"
        }
    }
    leftwall_common_stuff = nil

    -- TODO/FIXME: identical to floor except the texsurface_common (same attributes): optimize!
    local ceil_common_stuff = texsurface_common("ceil", "shaders/texsurface/texsurfaceX.vert")
    objects.opengl.ceil = {
        {
            type = "bind program",
            shader = ceil_common_stuff.programs
        },
        {
            -- uses the current shader
            type = "bind uniforms", -- exception? Multiple uniforms, since the single uniform case is unlikely
            uniforms = ceil_common_stuff.uniforms
        },
        {
            type = "bind attribute",
            name = "pos",
            values = {
                (-0.5), 0.5, 0,
                (-0.5), (-0.5), 0,
                0.5, (-0.5), 0,
                0.5, 0.5, 0
            },
            dim = 3
        },
        {
            type = "bind attribute",
            name = "uv",
            values = { 0, 1, 0, 0, 1, 0, 1, 1 },
            dim = 2
        },
        {
            -- good but not perfect, maybe. Should the index be put in a separate task?
            type = "draw",
            primitive = "GL_TRIANGLES", -- optional! GL_TRIANGLES is the default
            offset = 0, -- optional (default: 0)!
            --count  = 6, -- optional for indexed draws!
            elements = { 0, 1, 2, 2, 3, 0 }, -- optional! use draw elements otherwise
        },
        {
            -- no values -> unbind/disable attribute
            type = "bind attribute",
            name = "pos"
        },
        {
            -- no values -> unbind/disable attribute
            type = "bind attribute",
            name = "uv"
        }
    }
    ceil_common_stuff = nil

    local function getFrontWall()
        local front_common_stuff = texsurface_common("frontwall", "shaders/texsurface/texsurfaceX.vert")

        local uniforms = front_common_stuff.uniforms
        uniforms.model = nil
        uniforms.lightrange = nil
        uniforms.prevideo = nil
        uniforms.video = nil
        uniforms.video_transform = nil

        local result = {
            {
                type = "bind program",
                shader = front_common_stuff.programs
            },
            {
                -- uses the current shader
                type = "bind uniforms", -- exception? Multiple uniforms, since the single uniform case is unlikely
                uniforms = uniforms
            },
            {
                type = "bind attribute",
                name = "pos",
                values = {
                    (-0.5), 0.5, 0.0,
                    (-0.5), (-0.5), 0.0,
                    0.5, (-0.5), 0.0,
                    0.5, 0.5, 0.0,
                },
                dim = 3
            },
            {
                type = "bind attribute",
                name = "uv",
                values = { 0, 1, 0, 0, 1, 0, 1, 1 },
                dim = 2
            },
            {
                type = "draw",
                elements = { 0, 1, 2, 0, 2, 3 },
            },
        }
        front_common_stuff = nil
        return result
    end

    objects.opengl.frontwall = getFrontWall()
    local function getBackWall()
        local back_common_stuff = texsurface_common("frontwall", "shaders/texsurface/texsurfaceX.vert")

        local uniforms = back_common_stuff.uniforms
        uniforms.model = nil
        uniforms.lightrange = nil
        uniforms.prevideo = nil
        uniforms.video = nil
        uniforms.video_transform = nil

        local result = {
            {
                -- uses the current shader
                type = "bind uniforms", -- exception? Multiple uniforms, since the single uniform case is unlikely
                uniforms = uniforms
            },
            {
                type = "draw",
                elements = { 0, 1, 2, 0, 2, 3 },
            },
            {
                type = "bind attribute",
                name = "uv"
            },
            {
                type = "bind attribute",
                name = "pos"
            },
        }
        back_common_stuff = nil
        return result
    end

    objects.opengl.backwall = getBackWall()

    -- Temporarily removed TODO/FIXME re-add after the next release
    -- current.opengl.seats = require "scene/models/seats/seats"


    objects.opengl = functional.merge_maps(objects.opengl,
        scene_controls_opengl.generate_control_objects(screen_params.height * 0.45, true, false, parameters.ratio, parameters.stereo_type))

    objects.opengl.pointer = scene_controls_opengl.return_pointer(0.1,
        vr_config.get(options_consts.POINTER_OPACITY) / 100.0)

    local function get_subtitles_quad()
        local sub_width, sub_height = subtitles.compute_size(vr_config.get(options_consts.SUBTITLES_QUAD_DISTANCE),
            vr_config.get(options_consts.SUBTITLES_QUAD_SIZE) / 100.0,
            screen_points)

        return {
            {
                type = "bind program",
                shader = {
                    vert = "shaders/subtitles/subtitles.vert",
                    frag = "shaders/subtitles/subtitles.frag"
                }
            },
            {
                type = "bind uniforms",
                uniforms = {
                    projection = common_uniforms.projection,
                    modelview = common_uniforms.modelview,
                    subtitle = common_uniforms.subtitle,
                }
            },
            {
                type = "bind attribute",
                name = "pos",
                values = {
                    (-sub_width * 0.5), sub_height, 0,
                    (-sub_width * 0.5), 0, 0,
                    sub_width * 0.5, 0, 0,
                    sub_width * 0.5, sub_height, 0,
                },
                dim = 3
            },
            {
                type = "bind attribute",
                name = "uv",
                values = {
                    0, 0,
                    0, 1,
                    1, 1,
                    1, 0
                },
                dim = 2
            },
            {
                type = "draw",
                primitive = "GL_TRIANGLES",
                elements = { 0, 1, 2, 0, 2, 3 },
            },
            {
                type = "bind attribute",
                name = "pos"
            },
            {
                type = "bind attribute",
                name = "uv"
            }
        }
    end

    objects.opengl.subtitles_quad = get_subtitles_quad()
    objects.opengl.progress = opengl_common.progress_bar_model(common_uniforms)

    return objects
end

return {
    create_opengl_objects = create_opengl_objects,
}
