-- Functions for the creation of the standard theater or similar 
-- environments (like the cardboard theater, or the void one, at
-- least for now).
-- Created by Riccardo Di Meo
-- © Riccardo Di Meo 2015
local actions_consts = require "scene/actions_consts"
local consts = require "scene/scene_constants"
local functional = require "scene/functional"
local mymath = require "mymath"
local options = require "options"
local options_consts = require "scene/options_consts"
local quat = require "scene/quaternion"
local rendering = require "scene/theater/rendering"
local scene_controls = require "scene/controls/controls"
local screen = require "scene/theater/screen"
local subtitles = require "scene/subtitles"
local vr_config = require "scene/vr_config"

-- This is a sample of "global" debug geometry:
-- this object will be drawn for every visible and 
-- drawable object that doesn't specify a debugmodel 
-- itself, if the debug is turned on:
--debugmodel = "current.debug_objects.center"

local create = function(parameters)

    local screen_params = screen.get_right_screenmodel(parameters.ratio)

    local camera_pos = {
        0.0,
        -vr_config.get(options_consts.VIEWER_TO_FLOOR_DISTANCE) * options.room_height / 100.0, -- floor is at 0
        -(vr_config.get(options_consts.VIEWER_TO_SCREEN_DISTANCE) / 100.0 - 0.5) * options.room_depth, -- back is at -options.room_depth/2
    }

    local environment = vr_config.get(options_consts.THEATER_TYPE)

    ----------------
    -- scene objects
    local current = {}
    current.scene = {}
    current.scene.objects = {}
    current.scene.player = consts.CAMERA
    current.scene.locked = consts.SCREEN

    -- root node for the room
    current.scene.objects[#current.scene.objects + 1] = {
        name = consts.ROOM,
        position = { 0, 0, 0 },
        orientation = quat.front_up_to_quaternion({ 0, 0, -1 }, { 0, 1, 0 }),
    }
    -- Screen group
    current.scene.objects[#current.scene.objects + 1] = {
        name = consts.SCREENGROUP,
        position = { 0, 0, -options.room_depth * 0.5 + 0.01 },
        orientation = quat.front_up_to_quaternion({ 0, 0, -1 }, { 0, 1, 0 }),
        relativeto = consts.ROOM,
    }

    -- Position of the screen in respect to the screen group
    current.scene.objects[#current.scene.objects + 1] = {
        name = consts.SCREEN_PLACEMENT,
        position = { 0, screen_params.lift, 0 },
        orientation = quat.front_up_to_quaternion({ 0, 0, 1 }, { 0, 1, 0 }),
        relativeto = consts.SCREENGROUP,
    }
    -- The screen proper
    current.scene.objects[#current.scene.objects + 1] = {
        name = consts.SCREEN,
        position = { 0, 0, 0 },
        scale = { screen_params.width, screen_params.height, 1 },
        orientation = quat.front_up_to_quaternion({ 0, 0, 1 }, { 0, 1, 0 }),
        openglmodel = "current.opengl.screen",
        relativeto = consts.SCREEN_PLACEMENT,
        action = {
            trigger = {
                native = actions_consts.ACTION_ALIGN_AND_PAUSE, -- TODO/FIXME: not DRY!!!
            },
        }
    }
    local screen_points = {
        -- top left, bottom right
        { -screen_params.width * 0.5, screen_params.lift + screen_params.height * 0.5, (-options.room_depth * 0.5 + 0.01) },
        { screen_params.width * 0.5, screen_params.lift - screen_params.height * 0.5, (-options.room_depth * 0.5 + 0.01) },
    }

    -- TOFO/FIXME: eugh…
    -- save the index to inject some code as soon as we have the correct objects structure
    local screen_index = #current.scene.objects

    current.scene.objects[#current.scene.objects + 1] = scene_controls.create_seek_bar(
       screen_params,
       vr_config.get(options_consts.SEEK_BAR_SHOW_BAR),
       vr_config.get(options_consts.SEEK_BAR_POSITION))
    
    -- Position of the observer
    current.scene.objects[#current.scene.objects + 1] = {
        name = consts.CAMERA,
        position = camera_pos,
        orientation = quat.front_up_to_quaternion({ 0, 0, -1 }, { 0, 1, 0 }),
    }
    local camera_obj = current.scene.objects[#current.scene.objects]

    -- Nifty axes in the middle of the screen
    current.scene.objects[#current.scene.objects + 1] = {
        name = consts.AXES,
        visible = false,
        position = { 0, 0, 0 },
        orientation = quat.front_up_to_quaternion({ 0, 0, -1 }, { 0, 1, 0 }),
        openglmodel = "current.opengl.axes",
        relativeto = consts.ROOM,
    }
    if environment ~= options_consts.VOID_ENV then
        -- ROOM Walls
        current.scene.objects[#current.scene.objects + 1] = {
            name = consts.FLOOR,
            position = { 0, options.floor_level, 0 },
            orientation = quat.front_up_to_quaternion({ 0, 1, 0 }, { 0, 0, -1 }),
            scale = { options.room_width, options.room_depth, 1.0 },
            openglmodel = "current.opengl.floor",
            relativeto = consts.ROOM,
        }
        current.scene.objects[#current.scene.objects + 1] = {
            name = consts.CEIL,
            position = { 0.0, 0.0, options.room_height },
            orientation = quat.front_up_to_quaternion({ 0, 1, 0 }, { 0, 0, -1 }),
            scale = { options.room_width, options.room_depth, -1.0 },
            openglmodel = "current.opengl.ceil",
            relativeto = consts.ROOM,
        }
        current.scene.objects[#current.scene.objects + 1] = {
            name = consts.RIGHT_WALL,
            position = { 0.0, options.room_height * 0.5, options.room_width * 0.5 },
            orientation = quat.front_up_to_quaternion({ -1, 0, 0 }, { 0, 1, 0 }),
            scale = { options.room_depth, options.room_height, 1.0 },
            openglmodel = "current.opengl.rightwall",
            relativeto = consts.ROOM,
        }
        current.scene.objects[#current.scene.objects + 1] = {
            name = consts.LEFT_WALL,
            position = { 0.0, options.room_height * 0.5, -options.room_width * 0.5 },
            orientation = quat.front_up_to_quaternion({ -1, 0, 0 }, { 0, 1, 0 }),
            scale = { options.room_depth, options.room_height, -1.0 },
            openglmodel = "current.opengl.leftwall",
            relativeto = consts.ROOM,
        }
        current.scene.objects[#current.scene.objects + 1] = {
            name = consts.FRONT_WALL,
            position = { 0.0, options.room_height * 0.5, options.room_depth * 0.5 },
            orientation = quat.front_up_to_quaternion({ 0, 0, 1 }, { 0, 1, 0 }),
            scale = { options.room_width, options.room_height, 1.0 },
            openglmodel = "current.opengl.frontwall",
            relativeto = consts.ROOM,
        }
        current.scene.objects[#current.scene.objects + 1] = {
            name = consts.BACK_WALL,
            position = { 0.0, options.room_height * 0.5, options.room_depth * 0.5 },
            orientation = quat.front_up_to_quaternion({ 0, 0, -1 }, { 0, 1, 0 }),
            scale = { options.room_width, options.room_height, 1.0 },
            openglmodel = "current.opengl.backwall",
            relativeto = consts.ROOM,
        }
        --      current.scene.objects[#current.scene.objects + 1] = {
        --	 name = consts.SEATS,
        --         position = {0.0, 2.0, 0.0},
        --         orientation = quat.front_up_to_quaternion({0, 0, 1}, {0, 1, 0}),
        --         scale = {2.0, 2.0, 2.0},
        --         openglmodel = "current.opengl.seats"
        --      }
    end

    -- "resume" button on the screen
    current.scene.objects[#current.scene.objects + 1] = {
        name = consts.RESUME,
        visible = false,
        openglmodel = "current.opengl.resume",
        action = {
            trigger = {
                native = actions_consts.ACTION_RESUME,
            },
            vselected = {
                native = actions_consts.ACTION_RESUME,
            },
        },
        relativeto = consts.SCREEN_PLACEMENT,
        position = { 0, 0, 0.1 },
        orientation = quat.front_up_to_quaternion({ 0, 0, 1 }, { 0, 1, 0 }),
    }

    current.scene.objects[#current.scene.objects + 1] = {
        name = consts.SUBTITLES_BASE,
        position = subtitles.compute_position(vr_config.get(options_consts.SUBTITLES_QUAD_DISTANCE),
            screen_points,
            camera_obj),
        orientation = quat.front_up_to_quaternion({ 0, 0, -1 }, { 0, 1, 0 }),
        relativeto = consts.ROOM,
    }
    local subs_base = current.scene.objects[#current.scene.objects]

    current.scene.objects[#current.scene.objects + 1] = {
        name = consts.SUBTITLES,
        visible = parameters.show_subtitles,
        openglmodel = "current.opengl.subtitles_quad",
        relativeto = consts.SUBTITLES_BASE,
        position = { 0, 0, 0 },
        orientation = quat.front_up_to_quaternion({ 0, 0, -1 }, { 0, 1, 0 }),
        --      debugmodel = "current.debug_objects.axes"
    }

    local movement_function = function(displacement)
       camera_obj.position = mymath.add_3D(camera_obj.position, displacement)
       camera_obj.dirty = true
       subs_base.position = subtitles.compute_position(vr_config.get(options_consts.SUBTITLES_QUAD_DISTANCE),
                                                       screen_points,
                                                       camera_obj)
       subs_base.dirty = true
       return true
    end

    local config_save_function = function()
       local floor_distance = -camera_obj.position[2] * 100 / options.room_height
       vr_config.set(options_consts.VIEWER_TO_FLOOR_DISTANCE, floor_distance)
       assert(math.abs(vr_config.get(options_consts.VIEWER_TO_FLOOR_DISTANCE) - floor_distance))
       
       local screen_distance = 100.0 * (0.5 - camera_obj.position[3] / options.room_depth)
       vr_config.set(options_consts.VIEWER_TO_SCREEN_DISTANCE, screen_distance)
       
       assert(math.abs(vr_config.get(options_consts.VIEWER_TO_SCREEN_DISTANCE) - screen_distance))
    end

    local pitch = mymath.grad2rad(vr_config.get(options_consts.CONTROLS_PITCH))
    local yaw = mymath.grad2rad(vr_config.get(options_consts.CONTROLS_YAW))

    local extra_controls = scene_controls.create_theater_controls(
       camera_obj, options.ui, options.ui.buttons_radius,
       pitch, yaw,
       {
          movement = movement_function,
          zooming = nil,
          config_save = config_save_function,
    })

    current.scene.objects = functional.append_array(current.scene.objects, extra_controls)

    -- A dot used to highlight the light position
    --light = {
    --   name = "light",
    --   position = {0.0, 0.0, 0.0},
    --   front = {0, 0, -1},
    --   up = {0, 1, 0},
    --   openglmodel = "current.opengl.light"
    --}

    -- Activation feedback
    current.scene.objects[#current.scene.objects + 1] = {
        name = consts.PROGRESS,
        visible = false,
        openglmodel = "current.opengl.progress",
        position = { 0.0, 0.0, -6},
        orientation = quat.front_up_to_quaternion({ 0, 0, -1 }, { 0, 1, 0 }),
    }

    -- TODO/FIXME: this sucks sooo big time :(
    current.scene.objects[screen_index].action.start_focus = scene_controls.generate_screen_start_focus(current.scene.objects)
    current.scene.objects[screen_index].action.end_focus = scene_controls.generate_screen_end_focus(current.scene.objects)

    current.scene.generic_callbacks = scene_controls.generate_generic_callbacks(current.scene.objects)

    local objects = rendering.create_opengl_objects(parameters, screen_points, screen_params)

    current.opengl = objects.opengl

    return current
end

return {
    create = create
}
